<?php


// Contemplate Assembler (assembler.php) version 1.3.1
// Copyright 2010 Arlo Leach (http://www.arlomedia.com)
// This product includes software developed by Type A Learning Agency (http://www.typea.net)
// This file may be used, modified, or redistributed under the terms
// of the open-source agreement contained in docs/license.txt


// adjust some PHP configuration settings

	ini_set("pcre.backtrack_limit", 1000000000); // http://bugs.php.net/bug.php?id=40846 - this goes to 100x the default value for PHP versions before 5.2.0


// if the script is being used by the flattener, formulator or editor, we'll give a different kind of output

	$skip_rendering = (isset($_GET["skip_rendering"])) ? $_GET["skip_rendering"] : "" ;
	$formulator_fill = (isset($_GET["formulator_fill"])) ? $_GET["formulator_fill"] : "" ;
	$skip_processing = (isset($_GET["skip_processing"])) ? $_GET["skip_processing"] : "" ;
	$show_editor_links = (isset($_GET["show_editor_links"])) ? $_GET["show_editor_links"] : "" ;
	$editor_index = (isset($_GET["editor_index"])) ? $_GET["editor_index"] : "/meditate/" ;
	$settings_used = "skip_rendering,formulator_fill,skip_processing,show_editor_links,editor_index,"; // this prevents these values from being changed into arrays like other arguments
	

// set visible text here for easier localization

	$search_label_success = "Search results for";
	$search_label_failure = "Can't find";
	$search_label_none = "Please enter a search string";
	
	
// initialize variables that we might use before setting

	$error_log = "";
	$show_errors = "";
	$output = "";
	$output_two = "";
	$pre_render_files = 0;
	$read_page_definitions_from = "";
	$page_definitions = "";
	$offset = 0;
	$embed_variable_keys = Array();
	$embed_variable_values = Array();
	$embeds = Array();
	$tag_names = Array();
	$tag_names_flattened = "";
	$file_fields = "";
	$stored_selections = "";
	$latest_mod_date = 0;	
	$latest_mod_date_formatted = "";
	$visitor_ID = "";
	$template = "";
	$page_definition = "";
	$header = "";
	

// define some useful functions

	function get_file($file, $method, $check_last_modified) { // method can be "http", "filesystem", or "auto" (auto looks at $pre_render_files list)
		GLOBAL $output, $file_missing, $pre_render_files, $script_location;
		GLOBAL $template, $read_page_definitions_from, $latest_mod_date, $latest_mod_date_formatted;
		GLOBAL $check_URL;
		$text = "";
		if (preg_match("/\/(\.\w.*)/", $file)) { // don't read files whose names start with "."
			$text = show_error("Not authorized to read the hidden file named \"$matches[1]\"");
			$file_missing = 1;
		} else if (preg_match("/([^\w-\.\/\?\&\=\%])/", $file, $matches)) { // don't read files whose names contain characters that could be used to read remote files, or to hack into the open command (Perl only)
			$text = show_error("File names can't contain the character $matches[0]");
			$file_missing = 1;
		} else {
			
			// first, figure out if we're going to read over HTTP or through the filesystem...
			if (($method == "")||($method == "auto")) {
				if ($pre_render_files) {
					for ($n=0; $n < count($pre_render_files); $n++) {
						if (strstr($file, $pre_render_files[$n])) { $method = "http"; }
					}
				}
			}
			
			// ...then get the file
			if ($method == "http") {

				// 1) prepare the URL for this type of connection
				$file_URL = absolute_path($file, "http");

				// in PHP, the next three steps are the same for http or local connections, but it's easier to maintain the
				// different versions of this script if we lay out the conditional nesting the same way in every language
				// hence the repeated code here
				
				// 2) open a connection to the file (in PHP, steps 2, 3 and 5 are shared by the http and filesystem options, so the nesting is different here)
				@$myfile = fopen($file_URL, "r");
				if (! $myfile) {
					$text = show_error("Can't open the file named \"$file\" $check_URL");
					$file_missing = 1;
				} else {
				
					// 3) read the file
					while (!feof($myfile)) { $text .= fread($myfile, 4096); }
					$file_missing = 0;
					
					// 4) close the file connection
					fclose($myfile);
					
				}
	
				//$output .= "<p>reading $file_URL over http<p>";  // debug

			} else {

				// 1) prepare the URL for this type of connection
				$file_URL = absolute_path($file, "filesystem");
				
				// 2) open a connection to the file (in PHP, steps 2, 3 and 5 are shared by the http and filesystem options, so the nesting is different here)
				@$myfile = fopen($file_URL, "r");
				if (! $myfile) {
					$text = show_error("Can't open the file named \"$file\" $check_URL");
					$file_missing = 1;
				} else {
				
					// 3) read the file
					while (!feof($myfile)) { $text .= fread($myfile, 4096); }
					$file_missing = 0;
					
					// 4) for local files, catch the file's modification date (for $read_page_definitions_from and $template, plus content files when the template contains a env_last_modified tag)
					if ($check_last_modified) {
						$mtime = filemtime($file_URL);
						if ($mtime > $latest_mod_date) {
							$latest_mod_date = $mtime;
							$latest_mod_date_formatted = date("Y-m-d - g:i a", $mtime); // h for two-digit hour
						}
					}
	
					// 5) close the file connection
					fclose($myfile);
					
				}
	
				//$output .= "<p>reading $script_location_copy + $file_copy through filesystem<p>"; // debug

			}
				
			// finally, conform all line breaks to avoid cross-platform issues
			// we'd rather conform to Unix line breaks (\n), but that doesn't work in ASP, which doesn't support the s flag in regular expressions
			$text = str_replace("\r\n", "\r", $text); // change DOS to Mac
			$text = str_replace("\n", "\r", $text); // change Unix to Mac

		}
		return $text;
	}
	
	function get_row($source_text, $field_name, $file_type) {
		// this is like get_field except it returns the complete HTML or XML row rather than the contents of the row
		// this is used by the content editor to update content
		GLOBAL $output, $field_marker, $file_missing, $content_file;
		GLOBAL $check_URL, $check_field_label;
		if (stristr($file_type, "xml")) {
			
			$source_text = make_XML_paths($source_text);
			$tag_type = "content";
			if (preg_match("/(<".$tag_type."[^>]+name=[\"\']?".preg_quote($field_name, "/")."\b[^>]*>.*?<\/$tag_type>)/is", $source_text, $matches)) {
				$row = $matches[1];
			} else {
				$row = show_error("Can't find the $tag_type tag named \"$field_name\" in $content_file $check_URL");
			}

		} else {
			//print "parsing row $field_name<br>"; // debug backtrack_limit

			if (! strstr($field_name, $field_marker)) {  // but what if formatting gets between the marker and the label?
				$field_name = $field_marker.$field_name;  // in that case, remove these lines and change $content_field to $field_marker.*?$content_field below
			}
			// the following should work, but the .*? before $field_name is greedy unless we precede the whole expression with .*, in which case we run over the backtrack limit
			//if (preg_match("/(<tr[^>]*>.*?<td[^>]*>.*?$field_name\W.*?\/td>.*?<td[^>]*>.*<\/td>.*?<\/tr>)/is", $source_text, $matches)) {
			//	//print "okay<br>"; // debug backtrack_limit
			//	$row = $matches[1];  // do a rough trim
			//	$row = preg_replace("/(.*?".preg_quote($field_marker, "/").".*?)".preg_quote($field_marker, "/").".*/is", "$1", $row, 1);  // remove any rows after this one
			//	$row = preg_replace("/(.*<\/tr>).*/is", "$1", $row, 1);  // clean up after the previous trim
			//} else {
			if (strstr($source_text, $field_name)) {
				//print "okay<br>"; // debug backtrack_limit
				$source_text = preg_replace("/$field_name\w/", "$field_marker-dont-split-here-", $source_text); // block out any field names that begin with the target field name but don't match exactly
				$halves = explode($field_name, $source_text); // split the content before and after the field name
				$rough_trim = $field_name;
				if (strstr($halves[0], $field_marker)) {
					// if there are rows before the target row, join the content after the last preceding row
					$halves[0] = explode($field_marker, $halves[0]);
					$rough_trim = $halves[0][(count($halves[0]) - 1)].$rough_trim;
				} else {
					// otherwise just put the first half of the content back
					$rough_trim = $halves[0].$rough_trim;
				}
				if (strstr($halves[1], $field_marker)) {
					// if there are rows after the target row, join the content before the first subsequent row
					$halves[1] = explode($field_marker, $halves[1]);
					$rough_trim = $rough_trim.$halves[1][0];
				} else {
					// otherwise just put the second half of the content back
					$rough_trim = $rough_trim.$halves[1];
				}
				// now we can do something like our original regex to clean things up
				preg_match("/.*(<tr[^>]*>.*?<td[^>]*>.*?$field_name\W.*?\/td>.*?<td[^>]*>.*<\/td>.*?<\/tr>).*/is", $rough_trim, $matches);
				$row = $matches[1];
			} else {
				//print "preg error ".preg_last_error()."<br>"; // debug backtrack_limit
				if (! $file_missing) {
					$row = show_error("Can't find the field named \"$field_name\" in $content_file $check_URL $check_field_label");
				}
			}

		}
		return $row;
	}

	function update_row($row, $file_type, $new_name, $new_content) {
		// this takes a row, changes the content of the field and then returns the row code
		// this is used by the content editor to update content
		// this is easier to parse than an update_field function because here we're working with a row that is already separated from the larger content file
		GLOBAL $output, $field_marker, $file_missing, $content_file;
		GLOBAL $check_URL, $check_field_label, $line_ending;

		// conform all line breaks to avoid cross-platform issues
		// we have to do this because the content may be coming in from an external source like Meditate
		$new_content = str_replace("\r\n", "\r", $new_content); // change DOS to Mac
		$new_content = str_replace("\n", "\r", $new_content); // change Unix to Mac

		if (stristr($file_type, "xml")) {
			
			$row = make_XML_paths($row);
			$tag_type = "content";
			if (preg_match("/(<".$tag_type."[^>]+name=[\"\']?)".preg_quote($field_name, "/")."(\b[^>]*>).*(<\/$tag_type>)/is", $row, $matches)) {
				$row = $matches[1].$new_name.$matches[2].$new_content.$matches[3];
			} else {
				$row = show_error("Can't update the $tag_type tag named \"$new_name\" in $content_file $check_URL");
			}

		} else {		

			if (preg_match("/(<tr[^>]*>.*?<td[^>]*>.*?".preg_quote($field_marker, "/").")\w+(\W.*?\/td>.*?<td[^>]*>).*(<\/td>.*?<\/tr>)/is", $row, $matches)) {
				$row = $matches[1].$new_name.$matches[2].$new_content.$matches[3];
			} else {
				if (! $file_missing) {
					$row = show_error("Can't update the row named \"$new_name\" in $content_file $check_URL $check_field_label");
				}
			}

		}

		// change to the preferred line ending set in preferences.ini
		$output = str_replace("\r", $line_ending, $output);
	
		return $row;
	}
	
	function get_field_bg_color($source_text, $field_name, $file_type) {
		GLOBAL $output, $field_marker, $file_missing, $content_file;
		GLOBAL $check_URL, $check_field_label;
		if (stristr($file_type, "xml")) {
			
			$source_text = make_XML_paths($source_text);
			$tag_type = "content";
			if (preg_match("/<".$tag_type."[^>]+name=[\"\']?".preg_quote($field_name, "/")."\b[^>]*bgcolor=[\"\']?([\#\w]+)[\"\']?/is", $source_text, $matches)) {
				$bg_color = $matches[1];
			} else {
				$bg_color = "";
			}

		} else {		

			if (! strstr($field_name, $field_marker)) {  // but what if formatting gets between the marker and the label?
				$field_name = $field_marker.$field_name;  // in that case, remove these lines and change $content_field to $field_marker.*?$content_field below
			}
			if (preg_match("/".$field_name."[^\w<]*<\/td>[^<]*<td[^<]*bgcolor=[\"\']?([\#\w]+)[\"\']?/is", $source_text, $matches)) {
				$bg_color = $matches[1];
			} else {
				$bg_color = "";
			}

		}
		return $bg_color;
	}
	
	function get_field($source_text, $field_name, $file_type) {
		GLOBAL $output, $field_marker, $file_missing, $content_file;
		GLOBAL $check_URL, $check_field_label;
		if (stristr($file_type, "xml")) {
			
			$source_text = make_XML_paths($source_text);
			$tag_type = "content";
			if (preg_match("/<".$tag_type."[^>]+name=[\"\']?".preg_quote($field_name, "/")."\b[^>]*>(.*?)<\/$tag_type>/is", $source_text, $matches)) {
				$field_text = $matches[1];
			} else {
				$field_text = show_error("Can't find the $tag_type tag named \"$field_name\" in $content_file $check_URL");
			}

		} else {		

			if (! strstr($field_name, $field_marker)) {  // but what if formatting gets between the marker and the label?
				$field_name = $field_marker.$field_name;  // in that case, remove these lines and change $content_field to $field_marker.*?$content_field below
			}
			if (preg_match("/$field_name(\W.*\/td>.*<td[^>]*>.*<\/td>)/is", $source_text, $matches)) {
				$field_text = $matches[1];  // do a rough trim
				$field_text = preg_replace("/<\/td>/i", "", $field_text, 1);  // remove up to the previous </td>
				$field_text = preg_replace("/<td[^>]*>/is", "split-here-hey", $field_text, 1);  // mark the opening <td>
				$field_text = preg_replace("/".preg_quote($field_marker, "/")."/i", "split-here-hey", $field_text, 1);  // mark the next field marker
				$temp = explode("split-here-hey", $field_text);  // trim at the markers
				$field_text = $temp[1];
				if (preg_match("/(.*)<\/td>/is", $field_text, $matches)) {  // remove from the trailing </td>
					$field_text = $matches[1];
				}
			} else {
				if (! $file_missing) {
					$field_text = show_error("Can't find the field named \"$field_name\" in $content_file $check_URL $check_field_label");
				}
			}

		}
		return $field_text;
	}
	
	function get_form($source_text, $form_element) {
		GLOBAL $output;
		GLOBAL $check_URL, $content_file;
		$value = "";
		
		// this function doesn't recognize values like this: <input name = foo> we could make it do so by adding /s* around all the equals signs, but it doesn't now
		
		// before looking for the value, let's find out if it has single or double quotes around it or not
		if (preg_match("/name=[\"\']?$form_element\W[^>]*value=[^\"\']/is", $source_text, $matches)) { $value_match = "([^ >]*)"; }
		else if (preg_match("/name=[\"\']?$form_element\W[^>]*>.*?<option[^>]+value=[^\"\']/is", $source_text, $matches)) { $value_match = "([^ >]*)"; } // the previous line covers all other input types, but the select menu is different because the type and value are in different tags (we wouldn't need this if we could write a regex that spanned the <select> and <option> tags, but that wasn't too greedy with other input types)
		else if (preg_match("/name=[\"\']?$form_element\W[^>]*value=\'/is", $source_text, $matches)) { $value_match = "'([^\']*?)'"; }
		else if (preg_match("/name=[\"\']?$form_element\W[^>]*>.*?<option[^>]+value=\'/is", $source_text, $matches)) { $value_match = "'([^\']*?)'"; } // the previous line covers all other input types, but the select menu is different because the type and value are in different tags (we wouldn't need this if we could write a regex that spanned the <select> and <option> tags, but that wasn't too greedy with other input types)
		else { $value_match = "\"([^\"]*?)\""; } // we'll let the double quote be the default, so if something goes wrong we still might get the desired result
		
		// okay, now let's look for it
		if (preg_match("/<input[^>]+name=[\"\']?$form_element\W[^>]*type=[\"\']?text[\"\']?[^>]+value=$value_match"."[^>]*>/is", $source_text, $matches)) { $value = $matches[1]; }
		else if (preg_match("/<input[^>]+name=[\"\']?$form_element\W[^>]*type=[\"\']?radio[\"\']?[^>]+value=$value_match"."[^>]+checked[^>]*>/is", $source_text, $matches)) { $value = $matches[1]; }
		else if (preg_match("/<input[^>]+name=[\"\']?$form_element\W[^>]*type=[\"\']?checkbox[\"\']?[^>]+value=$value_match"."[^>]+checked[^>]*>/is", $source_text)) {
			// return a list (not array) of all the checked boxes
			while (preg_match("/<input[^>]+name=[\"\']?$form_element\W[^>]*type=[\"\']?checkbox[\"\']?[^>]+value=$value_match"."[^>]+checked[^>]*>/is", $source_text, $matches)) {
				$value .= "$matches[1], ";
				$source_text = preg_replace("/<input[^>]+name=[\"\']?$form_element\W[^>]*type=[\"\']?checkbox[\"\']?[^>]+value=$value_match"."[^>]+checked[^>]*>/is", "", $source_text, 1); // the 1 here is needed in PHP to make the replacement not global
			}
			$value = substr($value, 0, -2);
		}
		else if (preg_match("/<select[^>]+name=[\"\']?$form_element\W[^>]*>.*?<option[^>]+value=$value_match"."[^>]+selected[^>]*>.*?<\/select>/is", $source_text, $matches)) { $value = $matches[1]; }
		else if (preg_match("/<textarea[^>]+name=[\"\']?$form_element\W[^>]*>(.*?)<\/textarea>/is", $source_text, $matches)) { $value = $matches[1]; }
		else {
			if (preg_match("/name=[\"\']?$form_element\W/is", $source_text, $matches)) {
				// some HTML editors might remove the value attribute if it's empty, so maybe that's why we can't find the form element
				$value = "";
			} else {
				// otherwise, we give up
				$value = show_error("Can't find the form element named \"$form_element\" in $content_file $check_URL");
			}
		}
		
		return $value;
	}
	
	function get_setting($source_text, $setting, $allow_override, $allow_partial) {
		// allow most preferences to be overridden through the argument string
		GLOBAL $output, $settings_used;
		$value = "";
		$setting_no_spaces = $setting;
		$setting_no_spaces = preg_replace("/[ -]/", "_", $setting_no_spaces, -1);
		$settings_used .= "$setting_no_spaces,";
		
		$override = (isset($_GET[$setting_no_spaces])) ? $_GET[$setting_no_spaces] : "" ;
		if (($allow_override)&&($override != "")) {
			$value = $override;
			// don't allow users to pass code through the URL arguments
			$value = str_replace("<", "&lt;", $value);
			$value = str_replace(">", "&gt;", $value);
		} else {
			$source_text = "\r".$source_text; // this makes it easy to make sure settings start on a new line
			if ($allow_partial) {  // this lets us use more verbose labels in the ini file
				if (preg_match("/[^\w]".$setting."[^:\r]*: ([^\r]*)\r*/i", $source_text, $matches)) {
					$value = $matches[1];
				}
			} else {  // this ensures an exact match in the pages file
				if (preg_match("/[^\w]".$setting.": ([^\r]*)\r*/i", $source_text, $matches)) {
					$value = $matches[1];
				}
			}
		}
		if ($value == "0") { $value = 0; } // this ensures that a 0 value will evaluate false in every language
		return $value;
	}
	
	function get_page_name($page_definition, $source) { // look for a page definition that exactly matches a given dynamic URL
		GLOBAL $output;
        $source = "\r".$source; // this makes it easier to match a page definition on the first line of the file
		$page_name = (preg_match("/[^\w]([-\w]+)\:\s+".preg_quote($page_definition, "/")."[\&\s\r]/", $source, $matches)) ? $matches[1] : "" ;
		return $page_name;
	}
	
	function get_page_names_containing($args, $source) { // look for page definitions that contain a given set of arguments
		GLOBAL $output;
		$pages = Array();
		$source = "\r".$source; // this makes it easier to match a page definition on the first line of the file
		while (preg_match("/[^\w]([-\w]+)\:\s+[^\r]+=".preg_quote($args, "/")."[^\r]*[\&\s\r]/", $source, $matches)) {
			array_push($pages, $matches[1]);
			$source = preg_replace("/([^\w])[-\w]+\:\s+[^\r]+=".preg_quote($args, "/")."[^\r]*[\&\s\r]/", "$1", $source, 1);
		}
		return $pages;
	}
	
	function get_page_names($source) { // return a list of all the page names in a page definitions file
		GLOBAL $output;
		$page_names = Array();
		while (preg_match("/([-\w]+)\:\s+[\w]+/", $source, $matches)) {
			$this_page_name = $matches[1];
			array_push($page_names, $this_page_name);
			$source = str_replace($this_page_name.":", "", $source);
		}
		return $page_names;
	}
	
	function get_field_names($source, $file_type) { // return a list of all the content fields in a content file -- either HTML or XML depending on the second argument
		GLOBAL $output, $field_marker;
		$field_names = Array();

		if (stristr($file_type, "xml")) {

			$source = make_XML_paths($source);
			$tag_type = "content";
			while (preg_match("/<".$tag_type."[^>]+name=[\"\']?([-\w\/]+)[\"\']?[^>]*>/is", $source, $matches)) {
				$this_field_name = $matches[1];
				array_push($field_names, $this_field_name);
				$source = preg_replace("/<".$tag_type."[^>]+name=[\"\']?[-\w\/]+[\"\']?[^>]*>/is", "", $source, 1);
			}

		} else {

			while (preg_match("/".preg_quote($field_marker, "/")."([-\w]+)/", $source, $matches)) {
				$this_field_name = $matches[1];
				array_push($field_names, $this_field_name);
				$source = preg_replace("/".preg_quote($field_marker, "/")."/", "", $source, 1); // can't be str_replace because it must only replace once
			}

		}
		return $field_names;
	}
	
	function get_form_names($source) { // return a list of all the form elements in a content field
		GLOBAL $output;
		$form_names = Array();
		while (preg_match("/<input[^>]+name=[\"\']?([-\w]+)\W[^>]*>/is", $source, $matches)) {
			$this_form_name = $matches[1];
			array_push($form_names, $this_form_name);
			$source = preg_replace("/".preg_quote($matches[0], "/")."/", "", $source, 1); // we can't use str_replace here because we must only replace once
		}
		while (preg_match("/<select[^>]+name=[\"\']?([-\w]+)\W[^>]*>/is", $source, $matches)) {
			$this_form_name = $matches[1];
			array_push($form_names, $this_form_name);
			$source = preg_replace("/".preg_quote($matches[0], "/")."/", "", $source, 1); // we can't use str_replace here because we must only replace once
		}
		while (preg_match("/<textarea[^>]+name=[\"\']?([-\w]+)\W[^>]*>/is", $source, $matches)) {
			$this_form_name = $matches[1];
			array_push($form_names, $this_form_name);
			$source = preg_replace("/".preg_quote($matches[0], "/")."/", "", $source, 1); // we can't use str_replace here because we must only replace once
		}
		return $form_names;
	}
	
	function make_XML_paths($source) { // rewrite nested XML content as flat content with paths for tag names; assumes no more than one tag per line
		GLOBAL $output;
		$source = explode("\r", $source);
		$path = Array();
		for ($n=0; $n < count($source); $n++) {
			$this_line = $source[$n];
			if (preg_match("/<group[^>]+name=[\"\']?([-\w]+)[\"\']?[^>]*>/is", $this_line, $matches)) {
				$group_name = $matches[1];
				array_push($path, $group_name); // add the new group name to the path array
			} else if (strstr($this_line, "</group>")) {
				$group_name = array_pop($path); // remove the latest group name from the path array
			} else if (preg_match("/<content[^>]+name=[\"\']?([-\w]+)[\"\']?[^>]*>/is", $this_line, $matches)) {
				$field_name = $matches[1];
				if (count($path) > 0) {
					$current_path = implode("/", $path);
					$field_name = $current_path."/".$field_name;
					$this_line = preg_replace("/(<content[^>]+name=[\"\']?)[-\w]+([\"\']?[^>]*>)/is", "$1".$field_name."$2", $this_line, 1);
					$source[$n] = $this_line;
				}
			}
		}
		$source = implode("\r", $source);
		return $source;
	}
	
	function get_directory_listing($directory) { // return a list of all the items in the specified directory
		GLOBAL $output;
		if (file_exists(absolute_path($directory, "filesystem"))) {
			if ($dir = opendir(absolute_path($directory, "filesystem"))) {
				$files = Array();
				while ($file = readdir($dir)) {
					if (substr($file, 0, 1) != ".") { array_push($files, $file); } // that gets files that don't start with a dot
				}  
				closedir($dir);
			}
		}
		return $files;
	}
	
	function get_recursive_listing($start_folder, $recursive=0, $filter="") { // return a list of all the files in a given folder, including subfolders
		GLOBAL $output;
		$start_folder = preg_replace("/[\\/\\\]$/", "", $start_folder, 1); // remove a trailing slash
		$file_list = Array();
		$folder_list = Array($start_folder);
		while ($folder_list) {
			$this_folder_path = $folder_list[0];
			$this_folder_items = get_directory_listing($this_folder_path);
			for ($n=0; $n<count($this_folder_items); $n++) {
				$this_item_name = $this_folder_items[$n];
				$this_item_path = $this_folder_path."/".$this_item_name;
 				// if the item name contains a dot, we assume it's a file, if not, we assume it's a directory
 				if (! strstr($this_item_name, ".")) {
 					// we've found a folder; add it to the list of folders to scan
 					if ($recursive) {
						array_push($folder_list, $this_item_path);
					}
				} else if ((!$filter)||(strstr($this_item_name, $filter))) {
					// we've found a file; add it to the list of files to return
					$this_item_path = str_replace($start_folder."/", "", $this_item_path); // make results relative to the start folder, not to the site
					array_push($file_list, $this_item_path);
				}
			}
			array_shift($folder_list); // remove the folder we just scanned from the list
		}
		return $file_list;
	}
	
	function process_tag_nav($this_type) {
		GLOBAL $page_definitions, $page;
		
		// if we haven't already loaded the page definitions file, do that now
		if (! $page_definitions ) { $page_definitions = get_file($read_page_definitions_from[0], "filesystem", 1); }
		
		// if we came here from the expanded URL and not the page shortcut, get the page name now
		if (! $page) { $page = get_page_name($page_definition, $page_definitions); }

		$nav_page = $page;
		
		$nav_pages = Array();
		$nav_page_position = 0;
		$count_pages = 0;
		
		// find the block of pages (separated by double line breaks) that this page belongs to
		// first, split the page definitions file up into blocks
		$pages_blocks = "\r\r$page_definitions\r\r";
		$pages_blocks = str_replace("\r\r", "\r\r\r", $pages_blocks);
		if (preg_match("/.*\r\r(.*?\W".$nav_page.": .*?)\r\r.*/", $pages_blocks, $matches)) {
			// then find the block that contains this page
			$nav_block = $matches[1];
			$nav_block = preg_replace("/\:[ ]*[^\r]*/", "", $nav_block, -1);
			$nav_block = trim($nav_block);
			$nav_pages = explode("\r", $nav_block);
			$count_pages = count($nav_pages);
		}
		for ($j=0; $j<$count_pages; $j++) {
			// finally, parse through it to find this page's position in the block
			if (strtolower($nav_pages[$j]) == strtolower($nav_page)) {
				$nav_page_position = $j+1;
				$j = $count_pages; // stop the loop
			}
		}
					
		if (strstr($this_type, "nav_current")) {
			$content = round($nav_page_position); // return the position of the current page
		} else if (strstr($this_type, "nav_total")) {
			$content = $count_pages; // return the number of pages in this block
		} else if (strstr($this_type, "nav_first")) {
			$content = $nav_pages[0]; // return the name of the first page
		} else if (strstr($this_type, "nav_last")) {
			$content = $nav_pages[$count_pages - 1]; // return the name of the last page
		
		} else {

			// don't bother with the next part, if the current page doesn't exist
			if ($nav_page_position > 0) {
		
				if (strstr($this_type, "nav_next")) {
					$result_page_position = $nav_page_position + 1;
					if ($result_page_position > $count_pages) { $result_page_position = 1; }
					$content = $nav_pages[$result_page_position - 1]; // return the name of the requested page
				} else if (strstr($this_type, "nav_previous")) {
					$result_page_position = $nav_page_position - 1;
					if ($result_page_position < 1) { $result_page_position = $count_pages; }
					$content = $nav_pages[$result_page_position - 1]; // return the name of the requested page
				}
				
			} else {
				$content = show_error("Can't find the page named \"$nav_page\" in $read_page_definitions_from[0]");
			}
			
		}
		
		return $content;
	}
	
	function process_tag_env($this_type) {
		GLOBAL $latest_mod_date_formatted;
		preg_match("/env_(.+)/i", $this_type, $matches);
		$variable = $matches[1];
		if (stristr($variable, "last_mod")) {
			$content = $latest_mod_date_formatted;
		} else {
			$content = (isset($_SERVER[strtoupper($matches[1])])) ? $_SERVER[strtoupper($matches[1])] : "" ;
			$content = stripslashes($content); // these get escaped in PHP for some reason
		}
		return $content;
	}

	function show_error($message) {
		GLOBAL $output, $error_log, $show_errors, $formulator_fill;
		
		// prepare the shown message by removing quote marks
		$message_show = $message;
		$message_show = preg_replace("/\"/", "&quot;", $message_show, -1);
		$message_show = preg_replace("/'/", "&#039;", $message_show, -1);
		$message_show = "<table cellpadding=\"4\" border=\"1\"><tr><td class=\"error-text\"><b>Assembler error</b><br>$message_show</td></tr></table>";
		
		// prepare the hidden message by removing <br> tags
		$message_hide = $message_show;
		$message_hide = str_replace("<br>", ";", $message_hide);
		$message_hide = "<!-- Assembler error: $message_hide -->";
		
		// prepare the logged message by removing all but the first line
		$message_log = $message;
		if (strpos($message_log, "<br>")) {
			$message_log = substr($message_log, 0, strpos($message_log, "<br>"));
		}
		$message_log = strip_tags($message_log); // also remove any HTML formatting from it
		
		// add the log message to the ongoing error log
		if ($error_log) { $error_log .= ","; } // and comma-delimit multiple errors on a page
		$error_log .= $message_log;
		
		// add the shown and hidden messages to the output
		if (($show_errors)&&($formulator_fill != -1)) {
			// only display one error at a time, since subsequent errors are usually a result of the first one
			if (! strstr($output, "Assembler error")) {
				return $message_show;
			} else {
				return "";
			}
		} else {
			return $message_hide;
		}
		
	}
	
	function relative_path($path, $subfolder) { // takes a reference to a file within our site and changes the initial delimiters if necessary
		GLOBAL $output, $read_files_from, $read_files_from_count, $require_authentication, $require_authentication_count, $login, $read_files_recursively;
		$this_read_file = 0; $authenticated = 1; $error = ""; $validate_path = 0;
		$path = str_replace("://", "", $path);  // don't allow files from other servers
		if (strstr($path, "/")) { $validate_path = 1; } // we can skip the validation if we just have a file directly the content folder
		if ((substr($path,0,3) != "../")&&(substr($path,0,2) != "./")) {
			// if the path doesn't include a site subfolder, add the subfolder we passed in
			if ($subfolder) { $subfolder .= "/"; }
			$path = "../".$subfolder.$path;
		}
		if ($validate_path) {
			// allow for nonstandard paths, if they're listed in the ini file
			if ($read_files_from) {
				for ($n=0; $n < $read_files_from_count; $n++) {
					if ($read_files_recursively) {
						// allow matches to subfolders
						if (substr($path, 0, strlen($read_files_from[$n])) == $read_files_from[$n]) { $this_read_file = 1; }
					} else {
						// require an exact match to this folder
						$path_no_filename = preg_replace("/(.*\/)[^\/]*/", "$1", $path);
						if ($path_no_filename == $read_files_from[$n]) { $this_read_file = 1; }
					}
				}
			}
			//$output .= "<p>this_read_file: $this_read_file<p>";  // debug
			if (! $this_read_file) {
				if (preg_match("/(.*\/)[^\/]*/i", $path, $matches)) { $path_trimmed = $matches[1]; }
				$error = show_error("Not authorized to read from the path \"$path_trimmed\"<br>Hint: Add nonstandard paths to \"read files from\" list in preferences.ini");
			}
		}
		if ((! $error)&&($require_authentication)) {
			// see if we require authentication to view this file and show an error if necessary
			for ($n=0; $n<$require_authentication_count; $n++) {
				if (substr($path, 0, strlen($require_authentication[$n])) == $require_authentication[$n]) {
					$authenticated = ($login) ? 1 : 0 ;
					break;
				}
			}
			if (! $authenticated) {
				if (preg_match("/(.*\/)[^\/]*/i", $path, $matches)) { $path_trimmed = $matches[1]; }
				$error = show_error("Not authorized to read from the path \"$path_trimmed\"<br>Hint: Log in to \"require authorization to read files from\" list in preferences.ini");
				//header("HTTP/1.0 403 Not Authorized"); // how can we make this work, along with the 404 header above?
			}
		}
		if ($error) {
			return $error;
		} else {
			return $path;
		}
	}
	
	function absolute_path($path, $method) { // convert a reference to a file within our site to a complete filesystem path or absolute URL
		GLOBAL $output, $script_location, $request, $URL_rewriting_flag;

		if ($method == "filesystem") {
		
			if ($script_location == "NA") {
				$output .= show_error("Couldn't determine the filesystem path to this script. Please ensure that either the SCRIPT_FILENAME or PATH_TRANSLATED variable is available on your server.");
			} else {
				$script_location_copy = $script_location;
				$file_copy = $path;
				while (strstr($file_copy, "../")) {
					// this avoids results like contemplate/../templates/default.html
					$script_location_copy = preg_replace("/[^\/]+\/$/", "", $script_location_copy, 1); // remove the last directory name from the script location
					$file_copy = preg_replace("/\.\.\//", "", $file_copy, 1); // remove the first updelim from the filename (in PHP, we need to use preg_replace so we only do one replacement)
				}
				$path = $script_location_copy.$file_copy;
			}
			
		} else {
		
			if ($request == "NA") {
				$output .= show_error("Couldn't determine the web path to this script. Please ensure that either the SCRIPT_NAME or REQUEST_URI variable is available on your server.");
			} else {
				$thisserver = (isset($_SERVER["HTTP_HOST"])) ? $_SERVER["HTTP_HOST"] : "" ;
				$thispath = "";
				if ($URL_rewriting_flag) {
					if (preg_match("/\/([^=]*\/)$URL_rewriting_flag\//i", $request, $matches)) { $thispath = $matches[1]; }
				} else {
					if (preg_match("/\/([^=]*\/)contemplate\//i", $request, $matches)) { $thispath = $matches[1]; }
				}
				
				// pass along the authentication information if present
				$login = (isset($_SERVER["PHP_AUTH_USER"])) ? $_SERVER["PHP_AUTH_USER"].":".$_SERVER["PHP_AUTH_PW"]."@" : "" ; 
				
				$path = preg_replace("/\.\.\//", "http://$login$thisserver/$thispath", $path, 1); // replace the first updelim with the scheme and path (in PHP, we need to use preg_replace so we only do one replacement)
				$path = str_replace("../", "", $path); // remove additional updelims
			}
			
		}
		
		return $path;
	}

	function parse_int($str) { // remove all but numeric characters from a string and convert to an integer
		GLOBAL $output;
		$str = preg_replace("/\D/", "", $str, -1);
		$int = round($str);
		return $int;
	}

	function make_preview($before_str, $str, $after_str) { // return a little bit before and a little bit after the selected text (used for search results)
		GLOBAL $output, $search_results_preview_length;
		$ellipses_before = (strlen($before_str) > $search_results_preview_length) ? "..." : "" ;
		$ellipses_after = (strlen($after_str) > $search_results_preview_length) ? "..." : "" ;
		$before_str = substr($before_str, -(round($search_results_preview_length / 2)));
		$after_str = substr($after_str, 0, round($search_results_preview_length / 2));
		$before_str = preg_replace("/^\s+/", "", $before_str);
		$after_str = preg_replace("/\s+$/", "", $after_str);
		$preview = $ellipses_before.$before_str."<b>$str</b>".$after_str.$ellipses_after;
		return $preview;
	}

	function stri_replace($pattern, $replacement, $subject) { // do a case-insensitive replace; this is a substitute for a PHP function that doesn't exist!
		GLOBAL $output;
		return preg_replace("/".preg_quote($pattern, "/")."/is", $replacement, $subject, -1);
	}

	function implode_recursive($delimiter, $array) {
		$result = "";
		if (is_array($array)) {
			foreach ($array as $element) {
				if (is_array($element)) {
					$result .= implode_recursive($delimiter, $element);
				} else {
					$result .= $delimiter.$element;
				}
			}
		} else {
			$result = $array;
		}
		return $result;
	} 

	function random_string($strlen) {
		$chars = "0123456789bcdfghjklmnpqrstvwxyzBCDFGHJKLMNPQRSTVWXYZ";
		$str = "";
		for ($n=0; $n<$strlen; $n++) {
			$str .= substr($chars, rand(0, (strlen($chars) - 1)), 1);
		}
		return $str;
	}

// set up some error messages that we can use later

	$check_URL = "<br>Hint: Check your page definition";
	$check_embed = "<br>Hint: Check your embed tag";
	$check_field_label = "<br>Hint: Check your content file and remove any formatting tags between the field marker and the field label";
	$change_tag_name = "<br>Hint: Choose a different name for this embed tag";
	

// set up some primary variables

	// try to figure out the path to this script as it appears in the browser, before any URL rewriting
	// in IIS, this probably won't work, because REQUEST_URI will either not exist, or will show the path after rewriting, since rewriting is handled externally from IIS
	$request = "NA";
	if (isset($_SERVER["REQUEST_URI"])) {
		$request = $_SERVER["REQUEST_URI"];
		if (strpos($request, "?") !== false) {
			$request = substr($request, 0, strpos($request, "?"));
		}
	} else {
		// IIS doesn't always provide the REQUEST_URI variable
		if (isset($_SERVER["SCRIPT_NAME"])) {
			$request = $_SERVER["SCRIPT_NAME"];
		}
	}
	
	// if we know the browser path, figure out the part that triggers the rewrite; assume this is just one virtual directory name
	$URL_rewriting_flag = "";
	if (strpos($request, "contemplate/assembler") === false) {
		if (preg_match("/([^\/]*)\/[^\/]+$/i", $request, $matches)) {
			$URL_rewriting_flag = $matches[1];
		}
	}

	// figure out the full local path of the contemplate directory so we can build an absolute path that works on Windows computers
	// on Unix, this will include any symlinks used to construct the URL, so we can run the script from a symlink to the real contemplate folder, 
	// and paths will be relative to the apparent, rather than the actual, location of assembler.asp
	// on Windows, this will not include virtual directories, but constructing a path from a virtual directory wouldn't work anyway
	// because those aren't available at the file system level
	$script_location = "NA";
	if (isset($_SERVER["SCRIPT_FILENAME"])) {
		$script_location = $_SERVER["SCRIPT_FILENAME"];
	} else {
		// Windows doesn't provide the SCRIPT_FILENAME variable
		if (isset($_SERVER["PATH_TRANSLATED"])) {
			$script_location = $_SERVER["PATH_TRANSLATED"];
		}
	}
	$script_location = str_replace("\\", "/", $script_location);
	$script_location = str_replace("//", "/", $script_location); // not sure why PHP returns double slashes on Windows
	if (($URL_rewriting_flag)&&(preg_match("/(.*\/)$URL_rewriting_flag\//i", $script_location, $matches))) {
		$script_location = $matches[1]."contemplate/";
	} else {
		$script_location = preg_replace("/(.*\/)[^\/]+/", "$1", $script_location, 1);
	}


// read in a preferences (.ini) file if one exists, and show which preferences may be overridden from the URL string

	$preferences = get_file("data/preferences.ini", "filesystem", 0);
	if ($preferences) {
		$field_marker = get_setting($preferences, "field marker", 1, 1);
		$log_traffic = get_setting($preferences, "log site traffic", 1, 1);
		$show_errors = get_setting($preferences, "show errors", 1, 1);
		$pre_render_files = get_setting($preferences, "pre-render files", 1, 1);
		$read_page_definitions_from = get_setting($preferences, "read page definitions from", 1, 1);
		$read_files_from = get_setting($preferences, "read files from", 0, 1);
		$read_files_recursively = get_setting($preferences, "read files recursively", 0, 1);
		$require_authentication = get_setting($preferences, "require authentication", 0, 1);
		$prevent_page_caching = get_setting($preferences, "prevent page caching", 1, 1);
		$show_expanded_URLs = get_setting($preferences, "show expanded URLs", 1, 1);
		$split_page_names_on = get_setting($preferences, "split page names on", 1, 1);
		$external_link_target = get_setting($preferences, "external link target", 1, 1);
		$search_results_preview_length = get_setting($preferences, "preview length", 1, 1);
		$line_ending = get_setting($preferences, "line ending", 1, 1);
		$show_source = get_setting($preferences, "show source", 1, 1);
		$show_source_to = get_setting($preferences, "show source to", 0, 1);
		$host_lookups = get_setting($preferences, "host lookups", 1, 1);
		$track_visitors = get_setting($preferences, "track visitors", 1, 1);
	} else {
		$field_marker = "";
		$log_traffic = "";
		$show_errors = "";
		$pre_render_files = "";
		$read_page_definitions_from = "";
		$read_files_from = "";
		$read_files_recursively = "";
		$require_authentication = "";
		$prevent_page_caching = "";
		$show_expanded_URLs = "";
		$split_page_names_on = "";
		$external_link_target = "";
		$search_results_preview_length = "";
		$line_ending = "";
		$show_source = "";
		$show_source_to = "";
		$host_lookups = "";
		$track_visitors = "";
	}
	
	
// do a bit of processing on the settings, and set some defaults in case preferences.ini or individual settings are absent

	if ($field_marker == "") { $field_marker = "##"; }
	
	if ($show_errors === "") { $show_errors = "1"; }
	$show_errors = intval($show_errors);

	if (($pre_render_files)&&($pre_render_files != "1")&&($pre_render_files != "0")) {
		$pre_render_files = str_replace(", ", ",", $pre_render_files);
		$pre_render_files = explode(",", $pre_render_files);
	}
	
	//if ($read_page_definitions_from == "") { $read_page_definitions_from = "../contemplate/pages.txt"; }
	if ($read_page_definitions_from) {
		$read_page_definitions_from = str_replace(", ", ",", $read_page_definitions_from);
		$read_page_definitions_from = explode(",", $read_page_definitions_from);
	} else {
		$read_page_definitions_from = Array("data/pages.txt");
	}
	$read_page_definitions_from_count = count($read_page_definitions_from);

	if (! strstr($read_files_from, "../contemplate/")) { $read_files_from .= ", ../contemplate/"; }
	if (! strstr($read_files_from, "../contemplate/data/")) { $read_files_from .= ", ../contemplate/data/"; }
	if (! strstr($read_files_from, "../content/")) { $read_files_from .= ", ../content/"; }
	if (! strstr($read_files_from, "../templates/")) { $read_files_from .= ", ../templates/"; }
	if (preg_match("/^, (.*)/", $read_files_from, $matches)) { $read_files_from = $matches[1]; }
	if ($read_files_from) {
		$read_files_from = str_replace(", ", ",", $read_files_from);
		$read_files_from = explode(",", $read_files_from);
	} else {
		$read_files_from = Array();
	}
	$read_files_from_count = count($read_files_from);

	if ($read_files_recursively === "") { $read_files_recursively = "1"; }
	$read_files_recursively = intval($read_files_recursively);

	if ($require_authentication) {
		$require_authentication = str_replace(", ", ",", $require_authentication);
		$require_authentication = explode(",", $require_authentication);
	} else {
		$require_authentication = Array();
	}
	$require_authentication_count = count($require_authentication);

	if ($show_expanded_URLs === "") { $show_expanded_URLs = "0"; }
	$show_expanded_URLs = intval($show_expanded_URLs);

	if ($search_results_preview_length == "") { $search_results_preview_length = 128; }

	if ($line_ending == "") {
		$line_ending = "\n";
	} else {
		eval('$line_ending = '."\"$line_ending\";");
	}

	$show_source = intval($show_source);

	if ($show_source_to) {
		$show_source_to = str_replace(", ", ",", $show_source_to);
		$show_source_to = explode(",", $show_source_to);
		$show_source_to_count = count($show_source_to);
	}

	if ($host_lookups == "") { $host_lookups = "auto"; }

	if ($track_visitors === "") { $track_visitors = "0"; }
	$track_visitors = intval($track_visitors);


// check to see if the user is logged in via basic authentication

	$login = (isset($_SERVER["PHP_AUTH_USER"])) ? $_SERVER["PHP_AUTH_USER"].":".$_SERVER["PHP_AUTH_PW"] : "" ; 
					

// generate an SQL-standard timestamp

	$SQL_timestamp = date("Y-m-d H:i:s");

		
// if we're running as an include file for the sake of our functions only, stop processing now

	if ($skip_processing) {
		return;
	}


// write an HTTP header to prepare for text output

	$header .= "Content-type: text/html\r";
	if ($prevent_page_caching) {
  		// trick the browser into getting a new page every time
  		$header .= "Pragma: no-cache\r";
		$header .= "Cache-control: no-cache\r";
		$header .= "Expires: Mon, 07 May 1973 00:00:00 GMT\r";
	}
	if (($track_visitors)&&(!isset($_COOKIE["Contemplate"]))) {
		// if we want to track visitors and this visitor doesn't have a Contemplate cookie yet, generate a visitor ID and save it to the cookie (this assumes that if they have the cookie, it contains a visitor ID)
		$SQL_timestamp_stripped = preg_replace("/\D/", "", $SQL_timestamp);
		$visitor_ID = random_string(16)."-".$SQL_timestamp_stripped;
		$header .= "Set-Cookie: Contemplate=visitor_ID%3D$visitor_ID; Expires=".date("D, d M Y 00:00:00 \G\M\T", time("") + (60*60*24*365*10))."; Path=/; Version=1";
	}


// parse the URL to figure out the arguments for this page

	$page = (isset($_GET["page"])) ? $_GET["page"] : "" ;
	$requested_page = (isset($_GET["requested_page"])) ? $_GET["requested_page"] : "" ;
	if ($page) {
		// look in the page definitions file(s)...
		$page_definition = "";
		for ($i=0; $i<$read_page_definitions_from_count; $i++) {
			$this_read_page_definitions_from = $read_page_definitions_from[$i];
			$this_read_page_definitions_from = relative_path($this_read_page_definitions_from, "contemplate");
			if (stristr($this_read_page_definitions_from, "not authorized")) {
				$output .= $this_read_page_definitions_from;
			} else {
				$page_definitions = get_file($this_read_page_definitions_from, "filesystem", 1);
				if ($file_missing) {
					$output .= $page_definitions; // $page_definitions will contain the error message generated by get_file()
				} else {
					$page_definition = get_setting($page_definitions, $page, 0, 0);
					if ($page_definition) { // CHECK FOR THE PRESENCE OF THE TEMPLATE VALUE ELSEWHERE
						// let any real-time args tag along for the ride
						$more_args = (isset($_SERVER["QUERY_STRING"])) ? urldecode($_SERVER["QUERY_STRING"]) : "" ;
						$more_args = preg_replace("/&?page=[^&]+/", "", $more_args, 1);
						if ($more_args) { $page_definition .= "&".$more_args; }
						$page_definition = preg_replace("/&+/", "&", $page_definition, -1);
						if ($show_expanded_URLs) {
							// redirect to the page with the full URL visible
							$page_definition = stri_replace("&show_expanded_URLs=1", "", $page_definition, 1);
							$page_definition .= "&requested_page=".$page;
							header("Location: ../contemplate/assembler.php?".$page_definition);
							exit;
						}
						$i = $read_page_definitions_from_count;
					}
				}
			}
		}
		if ($page_definition == "") {
			$read_page_definitions_from_plural = ($read_page_definitions_from_count > 1) ? "s" : "" ;
			$output .= show_error("Can't find a definition for \"$page\" in the file$read_page_definitions_from_plural ".implode(", ", $read_page_definitions_from).".");
			//header("HTTP/1.0 404 Not Found"); // this seems to have the desired effect in IIS but not Apache
			//header("Status: 404 Not Found"); // is this redundant?
		}
	} else {
		// ...or just use the current URL
		$page_definition = (isset($_SERVER["QUERY_STRING"])) ? urldecode($_SERVER["QUERY_STRING"]) : "" ;
		if ($requested_page) { $page = $requested_page; }
	}
	
	// remove our own arguments from the page definition if present
	$page_definition = stri_replace("&formulator_fill=1", "", $page_definition, 1);
	$page_definition = stri_replace("&show_info=1", "", $page_definition, 1);
	
	// don't proceed from here unless we have a page definition
	if ($page_definition) {
	
		// create arrays from the argument values, so we can more easily refer to their individual parts
		$args = explode("&", $page_definition);
		$args_count = count($args);
		for ($i=0; $i<$args_count; $i++) {
			$pair = explode("=", $args[$i]);
			$key = $pair[0];
			$key = strtolower($key); // argument names are hereby declared case-insensitive
			// skip args that have the same name as a preference setting, so that we can control which settings can be overridden
			if (!strstr(",$settings_used,", ",$key,")) {
				$value = (count($pair) > 1) ? stripslashes($pair[1]) : "" ;
				// don't allow users to pass code through the URL arguments
				$value = str_replace("<", "&lt;", $value);
				$value = str_replace(">", "&gt;", $value);
				if (strstr($value, ",")) {
					${$key} = explode(",", $value);
				} else {
					if ((! strstr($key, "template"))&&(! strstr($key, "requested_page"))) {
						${$key} = Array();
						${$key}[0] = $value;
					} else {
						${$key} = $value;
					}
				}
			}
		}
		
	
// grab the text of the template file
	
		if ($template) {
			$template_text = relative_path($template, "templates");
			if (! stristr($template_text, "not authorized")) { // if there's an error, leave the error in the template_text variable to display it
				$template_text = get_file($template_text, "auto", 1);
			}
		} else {
			$template_text = "<!--#embed args -->"; // allow display of one piece of content directly, with no template
		}
		$output .= $template_text;
		
		
// make the core processing recursive (e.g. process embed tags in embedded content)

		$i = 1;
		while (preg_match("/<!--\#\s*embed.*-->/is", $output)||(preg_match("/<!--\#\s*if.*-->/is", $output))) {
		

// process any conditional tags to hide pieces of the template
	
			// first, find and number all if/else/end if sets
			$levels = Array();
			$greatest = 0;
			while (preg_match("/<!--\#\s*end[ ]*if\D/is", $output)) {
				// which is next, an if, else, or end if?
				if (preg_match("/<!--\#\s*if\D(.*)/is", $output, $matches)) { $length_after_if = strlen($matches[1]); } else { $length_after_if = -1; }
				if (preg_match("/<!--\#\s*else\D(.*)/is", $output, $matches)) { $length_after_else = strlen($matches[1]); } else { $length_after_else = -1; }
				if (preg_match("/<!--\#\s*end[ ]*if\D(.*)/is", $output, $matches)) { $length_after_end_if = strlen($matches[1]); } else { $length_after_end_if = -1; }
				if (($length_after_if > $length_after_end_if)&&($length_after_if > $length_after_else)) {
					$greatest++;
					array_push($levels, $greatest);
					$output = preg_replace("/<!--\#\s*if(\D)/i", "<!--#if$greatest$1", $output, 1);
					//$output .= "<br>adding if $greatest";  // debug
				} else if (($length_after_else > $length_after_if)&&($length_after_else > $length_after_end_if)) {
					$next = $levels[count($levels) - 1];
					$output = preg_replace("/<!--\#\s*else(\D)/i", "<!--#else$next$1", $output, 1);
					//$output .= "<br>adding else $next";  // debug
				} else {
					$next = array_pop($levels);
					$output = preg_replace("/<!--\#\s*end[ ]*if(\D)/i", "<!--#end if$next$1", $output, 1);
					//$output .= "<br>adding end if $next";  // debug
				}
			}
			
			// then evaluate and process each of them
			while (preg_match("/<!--\#\s*if(\d+)[ ]*\(([^\r<]*[^ ])\)[ ]*-->/is", $output, $matches)) {  // ignore extra spaces around "if" and expression
				$j = $matches[1];
				$this_expression = $matches[2];
				$native_expression = $this_expression;
				
				// the variable symbol $ is optional in the template; add it here if it's missing
				if (! preg_match('/\$[_\w]+\[/', $native_expression)) { $native_expression = preg_replace("/([_\w]+\[)/", "$"."$1", $native_expression, -1); }
				
				// these variables don't exist internally as arrays; this fakes it so they're still available in conditional tags
				if (($split_page_names_on != "")&&(preg_match('/\$page\\[\\d+\\]/i', $native_expression))) {
					$page_array = explode($split_page_names_on, $page);
					$native_expression = preg_replace('/\$page\[(\d+)\]/i', "\$page_array[$1]", $native_expression);
				} else {
					$native_expression = preg_replace('/\$page\[\d*\]/i', "\$page", $native_expression);
				}
				$native_expression = preg_replace('/\$template\[\d*\]/i', "\$template", $native_expression);
				$native_expression = preg_replace('/\$page_definition\[\d*\]/i', "\$page_definition", $native_expression);

				// if an array (other than the special variables above) doesn't have an index, return every element of the array as a string
				$native_expression = preg_replace('/\\$([_\\w]+)\\[\\]/i', "implode(\",\", \$$1)", $native_expression);
				
				// these keywords don't exist as arrays at all; this inserts their values as literal strings so they're available, too
				while (preg_match('/\$*(env_\w+)/i', $native_expression, $matches)) {
					$native_expression = preg_replace('/\$*env_\w+/i', "\"".process_tag_env($matches[1])."\"", $native_expression, 1);
				}
				while (preg_match('/\$*(nav_\w+)/i', $native_expression, $matches)) {
					$native_expression = preg_replace('/\$*nav_\w+/i', "\"".process_tag_nav($matches[1])."\"", $native_expression, 1);
				}

				// depending on the warning level, PHP shows errors if we try to test a variable that hasn't been set, so we have to find all the variables in the expression and test them in a safe way before evaluating it
				while (preg_match('/\$([_\w]+)\[([^\]])\]/i', $native_expression, $matches)) {
					if (isset(${$matches[1]})&&(count(${$matches[1]}) > $matches[2])) {
						$native_expression = str_replace("\$".$matches[1]."[".$matches[2]."]", $field_marker.$matches[1]."[".$matches[2]."]", $native_expression);
					} else {
						$native_expression = str_replace("\$".$matches[1]."[".$matches[2]."]", "\"\"", $native_expression);
					}
				}				
				$native_expression = str_replace($field_marker, "$", $native_expression);
				
				//$output .= $native_expression."<br>"; // debug
				eval('$this_evaluates = ('.$native_expression.');');
				if (preg_match("/<!--\#\s*if".$j."[ ]*\([^\r<]*[^ ]\)[ ]*-->(.*)<!--\#\s*end[ ]*if".$j."[ ]*-->/is", $output, $matches)) {
					// always show embed tags in the formulator
					$conditional_content = $matches[1];
					if ((preg_match("/<!--\#\s*embed/i", $conditional_content))&&($formulator_fill)) { $this_evaluates = 1; }
				}
				if (preg_match("/<!--\#\s*else".$j."\D/i", $output)) {
					if ($this_evaluates) {
						$output = preg_replace("/<!--\#\s*if".$j."[ ]*\(([^\r<]*[^ ])\)[ ]*-->/is", "", $output, 1); // remove the if tag...
						$output = preg_replace("/<!--\#\s*else".$j."\D.*end[ ]*if".$j."[ ]*-->/is", "", $output, 1); // ...and from the else tag to the end if tag
					} else {
						$output = preg_replace("/<!--\#\s*if".$j."\D.*else".$j."[ ]*-->/is", "", $output, 1); // remove from the if tag to the else tag...
						$output = preg_replace("/<!--\#\s*end[ ]*if".$j."[ ]*-->/is", "", $output, 1); // ...and the end if tag
					}
				} else {
					if ($this_evaluates) {
						$output = preg_replace("/<!--\#\s*if".$j."[ ]*\(([^\r<]*[^ ])\)[ ]*-->/is", "", $output, 1); // remove the if tag
						$output = preg_replace("/<!--\#\s*end[ ]*if".$j."[ ]*-->/is", "", $output, 1); // remove the end if tag
					} else {
						$output = preg_replace("/<!--\#\s*if".$j."\D.*end[ ]*if".$j."[ ]*-->/is", "", $output, 1); // remove from the if to the end if tag
					}
				}
			}
		

// find all the embed tags in the template and replace with temporary placeholders
		
			// slower: while (preg_match("/(.*?)<!--\#\s*embed\s*([^\r]*?[^ ])\s*-->(.*)/is", $output, $matches)) {  // remove extra spaces before embed type, around "embed", and after attributes
			while (preg_match("/(<!--\#\s*embed\s*)([^\r]*?[^ ])(\s*-->)/is", $output, $matches)) {  // ignore extra spaces before embed type, around "embed", and after attributes
			
				// slower: $before_embed = $matches[1]; $this_embed = $matches[2]; $after_embed = $matches[3];
				$this_embed_complete = $matches[1].$matches[2].$matches[3];
				$this_embed = $matches[2];
				$this_embed = strtolower($this_embed); // embed tags are hereby declared case-insensitive
				$this_embed_copy = $this_embed;
				$this_type = "";
				$this_embed_variable = "";
				$this_embed_strip = "";
				$this_embed_replace = Array();
				$this_embed_fill_defer = "";
				$this_embed_special = "";
				
				if (preg_match("/\s+variable/", $this_embed)) {  // look for a variable attribute, which tells us to save the embed content as a variable rather than add it to the output
					$this_embed_variable = 1;
					$this_embed = preg_replace("/\s+variable/", "", $this_embed, 1);
				}
				
				if (preg_match("/\s+strip/", $this_embed)) {  // look for a strip attribute, which tells us to remove all markup from the embed content
					$this_embed_strip = 1;
					$this_embed = preg_replace("/\s+strip/", "", $this_embed, 1);
				}
				
				if (preg_match("/\s+replace/", $this_embed)) {  // look for a replace attribute, which tells us to do some search-and-replaces on the embed content
					$this_replace_array = Array();
					while (preg_match("/\s+replace\s+\/([^\/]*)\/([^\/]*)\//", $this_embed, $matches)) {
						$search = $matches[1]; $replace = $matches[2];
						array_push($this_replace_array, $search);
						array_push($this_replace_array, $replace);
						$this_embed = preg_replace("/\s+replace\s+\/[^\/]*\/[^\/]*\//", "", $this_embed, 1);
					}
					$this_embed_replace = $this_replace_array;
					// $output .= "this_replace_array: ".$embeds_replace[$i - 1]; // debug
				}
				
				if (preg_match("/\s+fill_defer/", $this_embed)) {  // look for a fill_defer attribute, which tells us put the formulator fill fields at the end of the file
					$this_embed_fill_defer = 1;
					$this_embed = preg_replace("/\s+fill_defer/", "", $this_embed, 1);
				}
				
				$output = preg_replace("/<!--\#\s*embed\s*".preg_quote($this_embed_copy, "/")."\s*-->/i", "-embed-$i-", $output, -1); // this is global so that if the same embed tag appears more than once, we only have to parse it once
				//$output .= "<p>Found embed tag: $this_embed\r\r";  // debug
		
		
// process each embed tag and replace the placeholders with real content

			// show a warning if we have one of our special variable names; you can't name an embed tag with one of these names
		
				if (($this_embed == "page")||($this_embed == "template")||($this_embed == "page_definition")) {
					$content = show_error("The tag name \"$this_embed\" is a reserved word $change_tag_name");
					$this_type = "ignore";
				}
				
		
			// get the args if we've hard-coded them into the tag
		
				if (preg_match("/args[ ]*=[ ]*([^ ]*)/is", $this_embed, $matches)) {
					$this_args = $matches[1];
					while (preg_match("/(\w+)\[(\d*)\]/", $this_args, $matches)) {
						// process passthru/hard-coded combos
						$variable_name = $matches[1];
						$variable_index = $matches[2];
						// these variables don't exist internally as arrays; this fakes it so they're still available in passthru tags
						if (($variable_name == "page")||($variable_name == "template")||($variable_name == "page_definition")) {
							if (($split_page_names_on != "")&&($variable_name == "page")) {
								if ($variable_index == "") {
									// if no index is specified, return the whole page name
									$value = $page;
								} else {
									$page_array = explode($split_page_names_on, $page);
									$value = $page_array[$variable_index];
								}
							} else {
								$value = ${$variable_name};
							}
						} else {
							if ($variable_index == "") {
								// if no index is specified, return the whole array as a string
								$value = implode(",", ${$variable_name});
							} else {
								$value = ${$variable_name}[$variable_index];
							}
						}
						$key = $variable_name."[".$variable_index."]";
						$this_args = str_replace($key, $value, $this_args);
					}
					$this_args = preg_replace("/\+/", "", $this_args, -1);
					$this_embed = "hardcoded"; // it doesn't really matter what we use for the name here
					${$this_embed} = explode(",", $this_args);
					$this_embed_special = 1;
				}
				
		
			// set the type to passthru if we have a square bracket (and remove the optional $ to avoid errors)
		
				if (preg_match("/(\w+)\[(\d*)\]/", $this_embed, $matches)) {
					$this_embed = $matches[1];
					$this_embed_passthru_index = $matches[2];
					$this_type = "passthru";
					if (($this_embed == "page")||($this_embed == "template")||($this_embed == "page_definition")) {
						// treat these as hard-coded values that shouldn't be set in the formulator
						$this_embed_special = 1;
					}
				}

		
			// set the type to the name if we have a navigation or environment tag
		
				if ((stristr($this_embed, "nav_"))||(stristr($this_embed, "env_"))) {
					$this_type = $this_embed;
					$this_embed_special = 1;
				}
				
		
			// if we don't have any of the previously listed special types, the type is the first argument

				if (! $this_type) {
					if (isset(${$this_embed})&&(${$this_embed} != "")) {
						if (is_array(${$this_embed})) {
							$this_type = ${$this_embed}[0];
						} else {
							// we're dealing with a missing argument that happens to have the same name as an internal variable,
							// or an existing argument that gets overwritten by an internal variable
							// this code doesn't eliminate the need for the page/template/page_definition check above
							// it also doesn't avoid problems with conflicting internal variables that are arrays
							$content = show_error("The tag name \"$this_embed\" is a reserved word $change_tag_name");
							$this_type = "ignore";
						}
					} else {
						$this_type = "skip";
					}
				}
				
		
			// show popup menus rather than content if we're running from the formulator
		
				if ($formulator_fill) {
				
					// the first time through, strip any form tags from the template so they don't interfere with the popup menus
					if ($i == 1) {
						$output = preg_replace("/<form.*?".">/is", "", $output, 1);
						$output = preg_replace("/<\/form.*?".">/is", "", $output, 1);
					}
					
					// if this isn't a special kind of embed tag, make some menus, or an input box for passthru tags
					if (! $this_embed_special) {
					
						// don't display an input form for any tag more than once, even though it might appear in the template more than once
						$tag_names_flattened = implode(",", $tag_names);
						if (! strstr(",$tag_names_flattened,", ",$this_embed,")) {

							// if this is a passthru tag but a normal tag by this name appears later in the template, don't do anything now so we can fill out the argument using the menus
							if (!((strstr($this_type, "passthr"))&&(preg_match("/<!--\#\s*embed\s*$this_embed\s/is", $output)))) {
								if (! strstr($this_type, "passthr")) {
								
									$fill_forms = "
									
										<b>".strtoupper($this_embed)."</b>
										<a href=JavaScript:parent.inspect_content('$this_embed')
										 onmouseover='document.images.inspector_$this_embed.src=\"graphics/inspector_1.gif\"'
										 onmouseout='document.images.inspector_$this_embed.src=\"graphics/inspector_0.gif\"'><img 
										src=graphics/inspector_0.gif
										 width=12 height=12 hspace=2 border=0
										 name=inspector_$this_embed alt='Inspect'></a><br>
									
										<form name=contemplate_$this_embed>
									
											<select name=embed_type onchange=parent.set_type('$this_embed')>
												<option value=-1>Embed type...</option>
												<option value=file>File</option>
												<option value=field>Field</option>
												<option value=form>Form</option>
												<option value=random>Random</option>
												<option value=none>None</option>
											</select><br>
									
											<img src=../contemplate/graphics/spacer.gif width=1 height=2><br>
									
											<select name=embed_file onchange=parent.set_file('$this_embed')>
												<option value=0>(not needed)</option>
											</select><br>
									
											<img src=../contemplate/graphics/spacer.gif width=1 height=2><br>
									
											<select name=embed_field onchange=parent.set_field('$this_embed')>
												<option value=0>(not needed)</option>
											</select><br>
									
											<img src=../contemplate/graphics/spacer.gif width=1 height=2><br>
									
											<select name=embed_form>
												<option value=0>(not needed)</option>
											</select><p><br>
									
										</form>
										
									";
									
									if ($this_embed_fill_defer != 1) {
										// either display the popup menus right where the embedded content would be, and then stop parsing this tag...
										$content = $fill_forms;
										$this_embed_strip = "";
										$this_embed_replace = Array();
										$this_embed_variable = "";
										$this_type = "ignore";
									} else {
										// or go ahead and parse this tag in place, and show the fill menus at the end of the page
										$output = stri_replace("</html>", $fill_forms."</html>", $output);
									}
				
									if (isset(${$this_embed})&&(implode(",", ${$this_embed}) != "")) {
										$this_embed_flattened = implode(",", ${$this_embed});
										$stored_selections .= "parent.show_stored('$this_embed', '$this_embed_flattened');\r\t\t\t\t\t\t\t";				
									}
								
								} else {
								
									// process passthru values a little differently -- just show a text field for them
									
									$fill_forms = "
									
										<b>".strtoupper($this_embed)."</b><br>
									
										<form name=contemplate_$this_embed>
									
											<input type=text size=24 name=embed_value><br>
									
										</form>
									
									";
									
									// display the passthru field at the top of the page, because those tags are likely to appear inside HTML tags
									// this also emphasizes their global scope
									$output = preg_replace("/(<body[^>]*>)/is", "$1".$fill_forms, $output, 1);
					
									if (isset(${$this_embed})&&(implode(",", ${$this_embed}) != "")) {
										$this_embed_flattened = implode(",", ${$this_embed});
										$this_embed_flattened = str_replace("'", "\\'", $this_embed_flattened);
										$stored_selections .= "document.contemplate_$this_embed.embed_value.value = '$this_embed_flattened';\r\t\t\t\t\t\t\t";				
									}
				
								}
								
								array_push($tag_names, $this_embed); // keep a list of all the embed tags we processed
						
							}
						
						}
		
					}
					
				}
		
		
			// open the content file, which we can assume to be the second argument
			
				if (($this_type == "file")||(strstr($this_type, "field"))||(strstr($this_type, "random"))||(strstr($this_type, "form"))||(strstr($this_type, "menu"))) {
					$content_file = ${$this_embed}[1];
					$content_file = relative_path($content_file, "content");
					if (! stristr($content_file, "not authorized")) {
						$content = get_file($content_file, "auto", 1);
					} else {
						$content = $content_file;
						$this_type = "error";
					}
					if ($file_missing) { $this_type = "error"; }
				}
		
		
			// get the desired piece of content...
			
			// ...for file
		
				if ($this_type == "file") {
					// (no parsing necessary here)
				}
				
				
			// ...for field
				
				else if (strstr($this_type, "field")) {
					$file_type = ${$this_embed}[1]; // determines whether we parse as HTML or XML
					$content_field = ${$this_embed}[2];
					$content = get_field($content, $content_field, $file_type);
			   }
			   
			   
			// ...for random
		
				else if (strstr($this_type, "rand")) {
					$file_type = ${$this_embed}[1]; // determines whether we parse as HTML or XML
					$content_fields = get_field_names($content, $file_type);
					$last_field_index = count($content_fields) - 1;
					
					$content_field = $content_fields[rand(0, $last_field_index)];
					$content = get_field($content, $content_field, $file_type);       		
				}
				
		
			// ...for form
		
				else if (strstr($this_type, "form")) {
					$file_type = ${$this_embed}[1]; // determines whether we parse as HTML or XML
					$content_field = ${$this_embed}[2];
					$form_element = ${$this_embed}[3];
					$content = get_field($content, $content_field, $file_type);       		
					$content = get_form($content, $form_element);
				}
		
		
			// ...for search
				
				else if (strstr($this_type, "search")) {
					$search_string = (isset($_GET["search_string"])) ? $_GET["search_string"] : "" ;
					$search_files = (isset($_GET["search_files"])) ? $_GET["search_files"] : "" ;
					
					// skip if no search string entered
					if ($search_string != "") {
					
						// if we haven't already loaded the page definitions file, do that now
						if (! $page_definitions ) { $page_definitions = get_file($read_page_definitions_from[0], "filesystem", 1); }
					
						// loop through each content file
						$results = Array();
						$content_files = get_recursive_listing("../content", $read_files_recursively, $search_files);
						$content_files_count = count($content_files);
						for ($j=0; $j<$content_files_count; $j++) {
							$file_name = $content_files[$j];
							$file_text = get_file(relative_path($file_name, "content"), "filesystem", 0);
							//$output .= "Searching file ".relative_path($file_name, "content")."<p>\r\r"; // debug
			
							// if the file contains the search string...
							if (preg_match("/(.*?)(".preg_quote($search_string, "/").")(.*)/is", $file_text, $matches)) {
			
								// ...create a preview of the found text...
								$results_preview = make_preview($matches[1], $matches[2], $matches[3]);
								
								// ...and count how many times it appears in the field
								$frequency_file = 0;
							
								// then get the field names in the file, and loop through each field
								$content_fields = get_field_names($file_text, $file_name);
								$content_fields_count = count($content_fields);
								for ($k=0; $k<$content_fields_count; $k++) {
									$field_name = $content_fields[$k];
									$field_text = get_field($file_text, $field_name, $file_name);
									// we don't need to reinstate $field_name in PHP, because functions have their own variable scope
								
									// if the field contains the search string...
									$field_text = strip_tags($field_text); // exclude text contained in a tag
									if (preg_match("/(.*?)(".preg_quote($search_string, "/").")(.*)/is", $field_text, $matches)) {
										
										// ...create a preview of the found text...
										$results_preview = make_preview($matches[1], $matches[2], $matches[3]);
										
										// ...and count how many times it appears in the field
										$frequency_field = 0;
										while (stristr($field_text, $search_string)) {
											$frequency_field++;
											$frequency_file++;
											$field_text = preg_replace("/".preg_quote($search_string, "/")."/i", "", $field_text, 1); // we can't use str_replace here, because we only want to match once
										}
			
										// look up the relevant page names, then add the preview and frequency to our results array
										$found_args = "field,$file_name,$field_name";
										$pages = get_page_names_containing($found_args, $page_definitions);
										$pages_count = count($pages);
										for ($l=0; $l<$pages_count; $l++) {
											$results["$pages[$l] : $results_preview"] = $frequency_field;
										}
										
									}
								
								}
								
								// look up the relevant page names, then add the preview and frequency to our results array
								$found_args = "file,$file_name";
								$pages = get_page_names_containing($found_args, $page_definitions);
								$pages_count = count($pages);
								for ($l=0; $l<$pages_count; $l++) {
									$results["$pages[$l] : $results_preview"] = $frequency_file;
								}
								
							}
							
						}
						
						// rank the search results by frequency
						arsort($results); // unfortunately, this sorts by frequency in descending order, then not by name at all; so items of the same frequency will appear in a random order
						$results_keys = array_keys($results);
						$results_values = array_values($results);
						
						//$content = "<table cellspacing=8 cellpadding=0 border=0>\r";  // these lines provide a different display option
						$content = "";
						$keys_count = count($results_keys);
						for ($j=0; $j < $keys_count; $j++) {
							$this_page = substr($results_keys[$j], 0, strpos($results_keys[$j], " : "));
							$this_preview = substr($results_keys[$j], strpos($results_keys[$j], " : ") + 3);
							$this_frequency = $results_values[$j];
							//$content .= "<tr valign=top>\r<td class=body>$this_frequency</td>\r";
							//$content .= "<td class=body><a href=../contemplate/assembler.php?page=$this_page>contemplate/assembler.php?page=$this_page</a><br>\r<img src=../contemplate/graphics/spacer.gif width=1 height=4 alt=''><br>\r$this_preview\r</td>\r</tr>";
							if ($URL_rewriting_flag) {
								$content .= "<p class=\"search-results-preview\"><a href=\"../$URL_rewriting_flag/$this_page.html\" class=\"search-results-link\">$URL_rewriting_flag/$this_page.html</a><br>\r<img src=\"../contemplate/graphics/spacer.gif\" width=\"1\" height=\"4\" alt=\"\"><br>\r$this_preview ($this_frequency)</p>\r\r";
							} else {
								$content .= "<p class=\"search-results-preview\"><a href=\"../contemplate/assembler.php?page=$this_page\" class=\"search-results-link\">contemplate/assembler.php?page=$this_page</a><br>\r<img src=\"../contemplate/graphics/spacer.gif\" width=\"1\" height=\"4\" alt=\"\"><br>\r$this_preview ($this_frequency)</p>\r\r";
							}
						}
						//$content .= "</table>\r";
						
						if ($content) {
							$content = "<p class=\"search-results-heading\">$search_label_success \"$search_string\":</p>".$content;
						} else {
							$content = "<p class=\"search-results-heading\">$search_label_failure \"$search_string\"</p>";
						}
					} else {
						$content = "<p class=\"search-results-heading\">$search_label_none</p>";
					}
				}
		
		
			// ...for menu
			
				else if (strstr($this_type, "menu")) {
					$file_type = ${$this_embed}[1]; // determines whether we parse as HTML or XML
					$content_field = ${$this_embed}[2];
					$data_type = ${$this_embed}[3]; // text, graphics, or links
					$content = get_field($content, $content_field, $file_type);
					
					if ($data_type == "graphics") { $data_type = "images"; } // this gives backwards compatibility from versions 1.1.0 to 1.2.2

					if (!strstr($content, "Assembler error")) {
					
						if ($data_type != "html") { // for this type, we'll just return the complete content
					
							// strip out unnecessary tags and spaces to make this easier to parse
							$content = stri_replace("</li>", "", $content);
							$content = str_replace("\r", "", $content);
							$content = preg_replace("/\s\s/", " ", $content);
							
							// first, find and number all lists and items
							$levels = Array();
							$greatest = 0;
							$limit = 0; // this is to prevent runaways while developing; we can delete it later
							while ((preg_match("/<\/ul\D/is", $content))&&($limit < 100)) {
								// which is next, a list item or a new child list?
								if (preg_match("/<li\D(.*)/is", $content, $matches)) { $length_after_item = strlen($matches[1]); } else { $length_after_item = -1; }
								if (preg_match("/<ul\D(.*)/is", $content, $matches)) { $length_after_list = strlen($matches[1]); } else { $length_after_list = -1; }
								if (preg_match("/<\/ul\D(.*)/is", $content, $matches)) { $length_after_end_list = strlen($matches[1]); } else { $length_after_end_list = -1; }
								//if (preg_match("/<\/li>(.*)/is", $content, $matches)) { $length_after_end_item = strlen($matches[1]); } else { $length_after_end_item = -1; }
								if (($length_after_list > $length_after_item)&&($length_after_list > $length_after_end_list)) {
									$greatest++;
									array_push($levels, $greatest);
									$content = preg_replace("/<ul(\D)/i", "<ul$greatest$1", $content, 1);
									//$output .= "<br>adding list $greatest";  // debug
								} else if (($length_after_item > $length_after_list)&&($length_after_item > $length_after_end_list)) {
									$next = $levels[count($levels) - 1];
									$content = preg_replace("/<li(\D)/i", "<li$next$1", $content, 1);
									//$output .= "<br>adding item $next";  // debug
								} else {
									$next = array_pop($levels);
									$content = preg_replace("/<\/ul(\D)/i", "</ul$next$1", $content, 1);
									//$output .= "<br>adding end list $next";  // debug
								}
								$limit++;
							}
							
							// then work through each list from deepest to shallowest and create a string formatted like an array
							$child_lists = Array();
							$child_lists[0] = ""; // we won't use the 0 index, but pad it now for consistency across languages
							for ($j=$greatest; $j>0; $j--) {
								if (preg_match("/<ul".$j."[^>]*>(.*)<\/ul".$j.">/is", $content, $matches)) {
									$child_lists[$j] = $matches[1];
									$content = preg_replace("/<ul".$j."[^>]*>.*<\/ul".$j.">/is", "-child-list-$j-", $content);
									$child_lists[$j] = preg_split("/<li[^>]*>/", $child_lists[$j]);
									for ($k=1; $k<count($child_lists[$j]); $k++) {
										//$output .= "processing list $j item $k<br>"; // debug
										if (preg_match("/-child-list-(\d+)-/", $child_lists[$j][$k], $matches)) {
											$replace_index = $matches[1];
											$child_lists[$j][$k] = str_replace("-child-list-".$replace_index."-", "", $child_lists[$j][$k]);
										} else {
											$replace_index = 0;
										}
										if ($data_type == "text") {
											$child_lists[$j][$k] = strip_tags($child_lists[$j][$k]);
											$child_lists[$j][$k] = preg_replace("/([\[\]\'\"])/", "\\\\$1", $child_lists[$j][$k]); // escape characters that would break the array output
										}
										if ($data_type == "images") {
											if (preg_match("/src\s*=\s*[\"\']?([^\"\'\s>]+)[\"\'\s>]/i", $child_lists[$j][$k], $matches)) {
												$child_lists[$j][$k] = $matches[1];
											} else {
												$child_lists[$j][$k] = "";
											}
										}
										if ($data_type == "images_width") {
											if (preg_match("/width\s*=\s*[\"\']?([^\"\'\s>]+)[\"\'\s>]/i", $child_lists[$j][$k], $matches)) {
												$child_lists[$j][$k] = $matches[1];
											} else {
												$child_lists[$j][$k] = "";
											}
										}
										if ($data_type == "images_height") {
											if (preg_match("/height\s*=\s*[\"\']?([^\"\'\s>]+)[\"\'\s>]/i", $child_lists[$j][$k], $matches)) {
												$child_lists[$j][$k] = $matches[1];
											} else {
												$child_lists[$j][$k] = "";
											}
										}
										if ($data_type == "images_alt") {
											if (preg_match("/alt\s*=\s*[\"\']?([^\">]+)[\">]/i", $child_lists[$j][$k], $matches)) {
												$child_lists[$j][$k] = $matches[1];
												$child_lists[$j][$k] = preg_replace("/([^\\\\])'/", "$1\\'", $child_lists[$j][$k]); // escape any apostrophes in the alt text (unless they are already escaped)
											} else {
												$child_lists[$j][$k] = "";
											}
										}
										if (($data_type == "links")||($data_type == "active")) {
											if (preg_match("/href\s*=\s*[\"\']?([^\"\'\s>]+)[\"\'\s>]/i", $child_lists[$j][$k], $matches)) {
												$child_lists[$j][$k] = $matches[1];
											} else {
												$child_lists[$j][$k] = "";
											}
										}
										$child_lists[$j][$k] = "'".trim($child_lists[$j][$k])."'";
										if ($replace_index) {
											$child_lists[$j][$k] = preg_replace("/\\[/", "[".$child_lists[$j][$k].", ", $child_lists[$replace_index], 1);
										}
									}
									array_shift($child_lists[$j]); // the first element, anything before the first <li>, is junk, so throw it away
									$child_lists[$j] = "[".implode(", ", $child_lists[$j])."]";
								}
							}
							if (count($child_lists) > 1) {
								$content = $child_lists[1];
							} else {
								$content = "[]";
							}
							
							// if the data type is active, loop through the links array and find the active link (it would be nicer to catch this as we build the arrays, but at that point we don't know what position the active element will eventually hold in the multidimensional array)
							if ($data_type == "active") {
								$links = $content;
								$links = str_replace("[", "Array (", $links);
								$links = str_replace("]", ")", $links);
								eval('$links = '.$links.';');
								$array_path = "";
								for ($j=0; $j<count($links); $j++) {
									$element_flattened = implode_recursive(",", $links[$j]);
									//$output .= "checking element $j of ".count($links).": ".$element_flattened."<br>"; // debug
									if (preg_match("/\b$page\b/i", $element_flattened)) {
										//$output .= "element $j is a match<br>"; // debug
										$array_path .= $j;
										if (strstr($element_flattened, ",")) {
											$array_path .= ",";
											$links = $links[$j];
											$j = 0;
										} else {
											break; // we found it!
										}
									}
								}
								$content = "[".$array_path."]";
							}
							
						}
						
					}	
				}
		
		
			// ...for navigation
			
				else if ((strstr($this_type, "nav_current"))||(strstr($this_type, "nav_total"))||(strstr($this_type, "nav_next"))||(strstr($this_type, "nav_prev"))||(strstr($this_type, "nav_first"))||(strstr($this_type, "nav_last"))) {
					$content = process_tag_nav($this_type);						
				}
		
		
			// ...for environment variable
			
				else if (strstr($this_type, "env_")) {
					$content = process_tag_env($this_type);
										
					// uncomment this to see all available variables
					//while(list($key,$val)=each($_SERVER)) { // debug
					//	$output .= "$key = $val<br>";
					//}
				}
		
		
			// ...for passthrough/passthru
			
				else if (strstr($this_type, "passthr")) {
					// these variables don't exist internally as arrays; this fakes it so they're still available in passthru tags
					if (($this_embed == "page")||($this_embed == "template")||($this_embed == "page_definition")) {
						if (($split_page_names_on != "")&&($this_embed == "page")) {
							if ($this_embed_passthru_index == "") {
								// if no index is specified, return the whole page name
								$content = $page;
							} else {
								$page_array = explode($split_page_names_on, $page);
								if (count($page_array) > $this_embed_passthru_index) {
									$content = $page_array[$this_embed_passthru_index];
								} else {
									$content = "";
								}
							}
						} else {
							$content = ${$this_embed};
						}
					} else {
						if (isset(${$this_embed})) {
							if ($this_embed_passthru_index == "") {
								// if no index is specified, return the whole array as a string
								$content = implode(",", ${$this_embed});
							} else {
								if (count(${$this_embed}) > $this_embed_passthru_index) {
									$content = ${$this_embed}[$this_embed_passthru_index];
								} else {
									$content = "";
								}
							}
						} else {
							$content = "";
						}
					}
				}
		
		
			// ...or for nothing!
				
				else if (strstr($this_type, "skip")) {
					$content = "<!-- Embed tag \"$this_embed\" not parsed because of blank or too few arguments in URL -->";
				}
		
				else if (strstr($this_type, "ignore")) {
					// do nothing at all
				}
		
				else if (! strstr($this_type, "error")) {
					$content = show_error("Can't understand the data type \"$this_type\"<br>Available types: file, field, random, form, search, nav_current, nav_total, nav_next, nav_previous, nav_first, nav_last, env_*, or passthru $check_embed");
				}
				
				
			// skip the rest of the parsing if this tag generated an error
			
				if (!strstr($content, "Assembler error")) {
				
				
				// remove all markup from the embed content if the strip attribute was included in the embed tag
				
					if ($this_embed_strip) {
						$content = strip_tags($content);
					}
				   
				   
				// do some search-and-replaces if the replace attribute was included in the embed tag
				
					if ($this_embed_replace) {
						$replace_count = count($this_embed_replace) / 2;
						for ($j=0; $j < $replace_count; $j++) {
							$search = $this_embed_replace[$j*2];
							$replace = $this_embed_replace[($j*2)+1];
							$content = preg_replace("/$search/", "$replace", $content, -1);
						}
					}
				   
				   
				}
				
				
			// if requested, add shortcut links to the web editor pages
			
				if ($show_editor_links) {
					if ((strstr($this_type, "field"))||(strstr($this_type, "random"))||(strstr($this_type, "form"))||(strstr($this_type, "menu"))) {
						$editor_links = "";
						$editor_links .= "<p class=\"body editor-links\">";
						$editor_links .= "<a href=\"$editor_index?page=content_edit&path=".str_replace("../", "", $content_file)."&field=$content_field\">Edit</a>";
						$editor_links .= "&nbsp;&bull;&nbsp;";
						$editor_links .= "<a href=\"$editor_index?page=content_history&filter_content=".str_replace("../", "", $content_file).",$content_field\">History</a>";
						$editor_links .= "</p>";
						if ($this_embed_fill_defer != 1) {
							// either display the links right under the embedded content...
							$content .= $editor_links;
						} else {
							// or show the links at the end of the page, with a text label so we can see what they're for
							$label = ($content_field) ? $content_field : $content_file ;
							$output = stri_replace("</html>", "[".$label."]".$editor_links."</html>", $output);
						}
					}
				}
			

			// add the new content to the output...
			
				if (! $this_embed_variable) {
					
					// replace all instances normally, but only the first instance in the formulator
					if (!$formulator_fill) {
						$output = str_replace("-embed-$i-", $content, $output);
					} else {
						$output = preg_replace("/-embed-$i-/", $content, $output, 1);
						// delete the extra embed placeholders that we didn't use in the formulator
						$output = str_replace("-embed-$i-", "", $output);
					}
					
				} else {
				
				
			// ...or just save it as a variable and delete the embed placeholder
				
					$embed_variable_keys[count($embed_variable_keys)] = $this_embed;
					$embed_variable_values[count($embed_variable_values)] = $content;
					
					$output = str_replace("-embed-$i-", "", $output);
		
				}
				
		
			// finally, start over with the next embed
			
				$last_content = $content;
				$content = "";
				$i++;
			
			}
	
		}


// add formulator data to the bottom of the document (we can't do this until we're done with recursive parsing)

	if ($formulator_fill) {
	
		if (count($tag_names)) { // don't bother with all this processing if there are no fillable tags
		
			$content_folder_only = 1; // we'll tell JavaScript whether we're using any nonstandard file locations
			$content_files = Array();
			for ($j=0; $j<count($read_files_from); $j++) {
				$this_directory = $read_files_from[$j];
				if ((!strstr($this_directory, "../contemplate"))&&(!strstr($this_directory, "../templates"))) {
					if (!strstr($this_directory, "../content")) { $content_folder_only = 0; }
					$this_files = get_recursive_listing($this_directory, $read_files_recursively, "");
					for ($k=0; $k<count($this_files); $k++) {
						$this_file = $this_files[$k];
						array_push($content_files, $this_directory.$this_file);
					}
				}
			}
			$file_fields = "";
			$field_forms = "";
			
			// build an array of fields for each content file...
			for ($j=0; $j<count($content_files); $j++) {
				$this_file_name = $content_files[$j];
				$this_file_nopunctuation = $this_file_name;
				$this_file_nopunctuation = str_replace(".", "_", $this_file_nopunctuation);
				$this_file_nopunctuation = str_replace("/", "_", $this_file_nopunctuation);
				$this_file_nopunctuation = str_replace("-", "_", $this_file_nopunctuation);
				$this_file_nopunctuation = str_replace(" ", "_", $this_file_nopunctuation);
				$this_file_nopunctuation = preg_replace("/^_*(.*)_*$/", "$1", $this_file_nopunctuation, -1);
				$this_file_text = get_file($this_file_name, "filesystem", 0);
				$this_file_fields = get_field_names($this_file_text, $this_file_name);
				if (count($this_file_fields)) {
					$this_file_fields = implode("', '", $this_file_fields);
					$file_fields .= "parent.".$this_file_nopunctuation."_fields = Array('$this_file_fields'); // file $j\r\t\t\t\t\t\t\t";
				}
				
				// ...and an array of forms for each field
				if (count($this_file_fields)) { $this_file_fields = explode("', '", $this_file_fields); }
				for ($k=0; $k<count($this_file_fields); $k++) {
					$this_field_name = $this_file_fields[$k];
					$this_field_nopunctuation = $this_field_name;
					$this_field_nopunctuation = str_replace(".", "_", $this_field_nopunctuation);
					$this_field_nopunctuation = str_replace("/", "_", $this_field_nopunctuation);
					$this_field_nopunctuation = str_replace("-", "_", $this_field_nopunctuation);
					$this_field_nopunctuation = str_replace(" ", "_", $this_field_nopunctuation);
					$this_field_nopunctuation = preg_replace("/^_*(.*)_*$/", "$1", $this_field_nopunctuation, -1);
					$this_field_text = get_field($this_file_text, $this_field_name, $this_file_name);
					$this_field_forms = get_form_names($this_field_text);
					if (count($this_field_forms)) {
						$this_field_forms = implode("', '", $this_field_forms);
						$field_forms .= "parent.".$this_file_nopunctuation."_".$this_field_nopunctuation."_forms = Array('$this_field_forms'); // form $j:$k\r\t\t\t\t\t\t\t";
					}
				}
			}
			
			$tag_names_javascript = implode("', '", $tag_names);
			$content_files_javascript = implode("', '", $content_files);
	
			$fill_data = "
			
				<script language=JavaScript>
										
					function load_menu(menu_name, menu_title, values) {
						clear_menu(menu_name, menu_title);
						menu_name.options[0] = new Option(menu_title, '0');
						if (values.length) {
							for (i=0; i < values.length; i++) {
								this_value = values[i];
								if (menu_title == 'File...') {
									if (parent.content_folder_only) {
										// remove the '../content/' if all the files are in that folder
										this_value_display = this_value.substring(11, this_value.length);
									} else {
										// just remove the '../' otherwise
										this_value_display = this_value.substring(3, this_value.length);
									}
								} else {
									this_value_display = this_value;
								}
								menu_name.options[i+1] = new Option(this_value_display, this_value);
							}
						} else {
							menu_name.options[1] = new Option('(none available)', '0');
						}
					}
					
					function clear_menu(menu_name, menu_title) {
						menu_name.options[0] = new Option(menu_title, '0');
						for (i=menu_name.options.length; i > 1; i--) {
							menu_name.options[i-1] = null;
						}
					}
	
					parent.tag_types = Array('file', 'field', 'form', 'random', 'none');
					parent.tag_names = ('$tag_names_javascript') ? Array('$tag_names_javascript') : Array() ;
					parent.content_files = ('$content_files_javascript') ? Array('$content_files_javascript') : Array() ;
					
					parent.content_folder_only = $content_folder_only;
					
					$file_fields
					
					$field_forms
	
					$stored_selections
			
					parent.contemplate_toolbar.location = 'formulator/toolbar.html?step=4'; // the formulator waits for this code to reload the toolbar
					
				</script>
				
			";
			
		} else {
		
			// if there are no fillable tags, advance the formulator toolbar to the next step
			$fill_data = "
				<script language=JavaScript>
					parent.tag_names = Array();
					parent.contemplate_toolbar.location = 'formulator/toolbar.html?step=5'; // the formulator waits for this code to reload the toolbar
				</script>
			"; 
			
		}
		
		$output = stri_replace("</html>", $fill_data."</html>", $output); // insert the JavaScript at the end of the file
			
	}
			
						
// merge argument values back together, since other scripts won't expect them to be arrays
	
		for ($i=0; $i<$args_count; $i++) {
			$pair = explode("=", $args[$i]);
			$key = $pair[0];
			if (!strstr(",$settings_used,", ",$key,")) {
				$value = (count($pair) > 1) ? $pair[1] : "" ;
				${$key} = $value;
			}
		}
	
	
// put content into any embed tag names marked with the variable attribute (we do this after merging argument variables in case they have the same names)
// first we'll rename the "hardcoded" variables if we have more than one

		if (count($embed_variable_keys)) {
			$hardcoded_count = 0;
			$hardcoded_index = 1;
			for ($i=0; $i<count($embed_variable_keys); $i++) {
				if ($embed_variable_keys[$i] == "hardcoded") {
					$hardcoded_count++;
				}
			}
			for ($i=0; $i<count($embed_variable_keys); $i++) {
				if ($hardcoded_count > 1) {
					$embed_variable_keys[$i] = "hardcoded_".$hardcoded_index;
					$hardcoded_index++;
				}
				${$embed_variable_keys[$i]} = $embed_variable_values[$i];
			}
		}
	
	
// set external link targets if requested in prefs file
	
		if ($external_link_target) {
			// remove any quotes from the external link target, and we'll make it match the quoting style of the href value
			$external_link_target = preg_replace("/[\"']/", "", $external_link_target);
			// temporarily remove server-side script tags that might confuse this parsing
			$output = preg_replace("/<\?(.*?)\?>/", "script_tag_start_!@#$1script_tag_end_!@#", $output, -1);
			// add our default target to every anchor tag
			$output = preg_replace("/<a href=(\\\?[\'\"]?)(\w+:\/\/[^>]*)>/", "<a href=$1$2 target=$1$external_link_target$1>", $output, -1);
			// then remove it from every tag that already had one
			$output = preg_replace("/<a href=(\\\?[\'\"]?)(\w+:\/\/[^>]*target=[^>]*) target=[^>]*>/", "<a href=$1$2>", $output, -1);
			// restore server-side script tags
			$output = preg_replace("/script_tag_start_!@#(.*?)script_tag_end_!@#/", "<"."?$1?".">", $output, -1);
		}


// if we're showing editor links, propagate that request to all the relative links on the page
// also include a return link at the top of the page
	
		if ($show_editor_links) {
			// update URLs like assembled/page.html?foo=bar
			$output = preg_replace("/(['\"][^:\?\&\s'\">]+\.html?\?[^\s'\">]+)/i", "$1&show_editor_links=1&editor_index=$editor_index", $output, -1);
			// update URLs like assembled/page.html
			$output = preg_replace("/(['\"][^:\?\&\s'\">]+\.html?)([^\?])/i", "$1?show_editor_links=1&editor_index=$editor_index$2", $output, -1);
			// update URLs like contemplate/assembler.php?page=page or contemplate/assembler.php?template=default&main=field,main,home
			$output = preg_replace("/(contemplate\/assembler\.php\?[^\s'\">]+)/i", "$1&show_editor_links=1&editor_index=$editor_index", $output, -1);
		
			// add a return link right after the body tag
			$return_link = "";
			$return_link .= "<div style=\"background-color: #666666; border-bottom: solid #333333 1px; padding: 8px;\">";
			$return_link .= "<a href=\"$editor_index?page=content_edit\" style=\"font-family: Arial; font-size: 13px; font-weight:  bold; color:  #FFFFFF; margin: 0px;\">Return to editor</a>";
			$return_link .= "</div>";
			$output = preg_replace("/(<body[^>]*>\s*)/i", "$1\n$return_link", $output, 1);
		}


	} // end of the "if $page_definition" check


// return the parsed text as an HTML file

	$header_array = explode("\r", $header);
	for ($i=0; $i<count($header_array); $i++) {
		header($header_array[$i]);
	}	
	
	// change to the preferred line ending set in preferences.ini
	$output = str_replace("\r", $line_ending, $output);

	// evaluate the output as PHP, unless the flattener tells us to skip it
	if ($skip_rendering) {
		print $output;
	} else {
		eval("?".">".$output); // this returns the output while rendering the code it contains
	}
	
	// if requested, display the output in a numbered, pre-rendered format so we can see where an error occurred
	// but only allow this from IP addresses listed in the preferences file
	$show_source_approved = 0;
	if ($show_source_to) {
		for ($i=0; $i < $show_source_to_count; $i++) {
			if ((isset($_SERVER["REMOTE_ADDR"]))&&($_SERVER["REMOTE_ADDR"] == $show_source_to[$i])) { $show_source_approved = 1; }
		}
	}
	if (($show_source)&&($show_source_approved)) {
		$output_numbered = $output;
		$output_numbered = str_replace("<", "&lt;", $output_numbered);
		$output_numbered = str_replace(">", "&gt;", $output_numbered);
		$output_numbered = explode($line_ending, $output_numbered);
		for ($i=0; $i<count($output_numbered); $i++) {
			$output_numbered[$i] = "<tr valign=top><td><font face=geneva,verdana,sans-serif color=#999999>".($i+1)."</font></td><td>&nbsp;</td><td><font face=geneva,verdana,sans-serif>".$output_numbered[$i]."</font></td></tr>";
		}
		$output_numbered = implode($line_ending, $output_numbered);
		print "<html>\r<body bgcolor=#FFFFFF>\r<p>\r<table>\r$output_numbered\r</table>\r</p>\r</body>\r</html>\r";
	}


// log this hit in the traffic reports

	if (($log_traffic)&&(!$skip_rendering)&&(!$formulator_fill)) {
	
		$status = (isset($_SERVER["REDIRECT_STATUS"])) ? $_SERVER["REDIRECT_STATUS"] : "" ;
	
		if ((strstr($request, "contemplate/assembler"))||($URL_rewriting_flag)||($status == "200")) {
			// if we recognize the request as a call to the assembler, display either the page name or the expanded URL
			if (!$page) { $page = $page_definition; }
		} else {
		 	// ... otherwise, we must have been redirected to here, so show the original request with the error code
		 	$page = substr($request,1); // the substr removes the initial slash
		 	$error_log = ($status) ? $status : "Unknown" ;
		}
		
		// catch the other environment variables now
		$IP_address = (isset($_SERVER["REMOTE_ADDR"])) ? $_SERVER["REMOTE_ADDR"] : "" ;
		$host = (isset($_SERVER["REMOTE_HOST"])) ? $_SERVER["REMOTE_HOST"] : $IP_address ; // if we knew that all servers substituted the IP address for us when the host is missing, then we could remove the conditional here
		$user_agent = (isset($_SERVER["HTTP_USER_AGENT"])) ? $_SERVER["HTTP_USER_AGENT"] : "" ;
		$referrer = (isset($_SERVER["HTTP_REFERER"])) ? $_SERVER["HTTP_REFERER"] : "" ;
		
		// find out whether we need to do our own host lookup
		if ($host_lookups == "auto") {
			$host_lookups = ($host == $IP_address) ? "on" : "off" ;
		}
		if ($host_lookups == "on") {
			$host = gethostbyaddr($IP_address);
		}

		// add the visitor ID to each entry if we're using cookies to track visitors
		if ($track_visitors) {
			if ($visitor_ID == "") {
				// get the visitor ID from the cookie if we didn't just generate one on this page
				// if we ever use the cookie to store other data, we can move this code to the top the script so that it's available to the rest of the script; right now it's here so that we can only run it when we need it
				$cookie_data = Array();
				$cookie_data_raw = $_COOKIE["Contemplate"];
				$cookie_data_raw = explode("&", $cookie_data_raw);
				for ($i=0; $i<count($cookie_data_raw); $i++) {
					$cookie_data_item = explode("=", $cookie_data_raw[$i]);
					$cookie_data[$cookie_data_item[0]] = $cookie_data_item[1];
				}
				$visitor_ID = $cookie_data["visitor_ID"];
			}
			$visitor_ID_with_tab = "$visitor_ID\t";
		} else {
			$visitor_ID_with_tab = "";
		}
		
		$SQL_timestamp_split = explode("-", $SQL_timestamp);
		$year = $SQL_timestamp_split[0];
		$month = $SQL_timestamp_split[1];
		$traffic_file = "data/reporter/".$year."-".$month.".txt";
		$error_file = "data/reporter/".$year."-".$month."-errors.txt";
		// leave the next two lines commented out to use relative paths and avoid a bug with PHP and open_basedir before version 4.3
		//$traffic_file = absolute_path($traffic_file, "filesystem");
		//$error_file = absolute_path($error_file, "filesystem");
		
		if (!$error_log) {
			$myfile = @fopen($traffic_file, "a");
			if (! $myfile) {
				$output_two .= "<!-- Can't open the traffic data file $traffic_file -->\r";
			} else {
				fputs($myfile, "$SQL_timestamp\t$visitor_ID_with_tab$IP_address\t$host\t$page\t$user_agent\t$referrer\r");
				fclose($myfile);
			}
		} else {
			$myfile = @fopen($error_file, "a");
			if (! $myfile) {
				$output_two .= "<!-- Can't open the error file $error_file -->\r";
			} else {
				fputs($myfile, "$SQL_timestamp\t$visitor_ID_with_tab$IP_address\t$host\t$page\t$user_agent\t$referrer\t$error_log\r");
				fclose($myfile);
			}
		}
		
		print $output_two;

	}

?>