<?php

// Contemplate Flattener (flattener/flattener.php) version 1.3.0
// Copyright 2009 Arlo Leach (http://www.arlomedia.com)
// This product includes software developed by Type A Learning Agency (http://www.typea.net)
// This file may be used, modified, or redistributed under the terms
// of the open-source agreement contained in docs/license.txt


// initialize variables that we might use before setting

	$flatten = 0;
	$input_file = "";
	$output_file = "";
	$stop = 0;
	$result = "";
	$output = "";
	$page_name = "";
	$pre_render_files = 1;
	$this_failure = 0;
	$changed = 0;
	$field_marker = "";
	$read_page_definitions_from = "";
	$unconvert_error = "";
	$changed_files = "";
	$check_URL = "";
	$settings_used = "";
	$template = "";
	$latest_mod_date = 0;
	

// define some useful subroutines
	
	function convert_URLs($text,$name) {
		GLOBAL $result, $URL_rewriting_flag_default;

		// convert URLs like ../assembled/page.html
		$text = preg_replace("/$URL_rewriting_flag_default\//i", "flattened/", $text, -1);
		
		// convert URLs like ../contemplate/assembler.php?page=page
		$text = preg_replace("/contemplate\/assembler\.php\?page\=([^<>\s\'\"]+)([<>\s\'\"])/i", "flattened/$1.html$2", $text, -1);
		
		// convert URLs like ../contemplate/assembler.php?template=default&main=field,main,home
		if (preg_match("/contemplate\/assembler\.php\?(template\=[^<>\s\'\"]+)[<>\s\'\"]/i", $text, $matches)) {
			$dynamic_page = $matches[1];
			$page_name = get_page_name($dynamic_page);
			if ($page_name != "") {
				$text = preg_replace("/contemplate\/assembler\/.php\?(template\=[^<>\s\'\"]+)([<>\s\'\"])/i", "flattened\/$page_name\.html$2", $text, -1);
			} else {
				$result .= "<font color=red>Error: Can't find a page name for $dynamic_page in $name.</font><br>";
			}
		}
	
		return $text;
	}

	function unconvert_URLs($text,$format_sample) {
		GLOBAL $unconvert_error, $URL_rewriting_flag_default;
		$flat_path = "../flattened/";
		$dynamic_path_short = "../$URL_rewriting_flag_default/";
		$dynamic_path_medium = "../contemplate/assembler.php?page=";
		if ((stristr($format_sample, $dynamic_path_short))&&(stristr($format_sample, $dynamic_path_medium))) {
			$unconvert_error = "Error: Assembled page may only contain short or medium format dynamic URLs, not both<br>";
		} else {
			if (stristr($format_sample, $dynamic_path_short)) {
				// convert flattened URLs to ../assembled/page.html
				$text = stri_replace($flat_path, $dynamic_path_short, $text);
			} elseif (stristr($format_sample, $dynamic_path_medium)) {
				// convert flattened URLs to ../contemplate/assembler.php?page=page
				$text = preg_replace("/".preg_quote($flat_path, "/")."([^<>\s\'\"]+)\.html/i", $dynamic_path_medium."$1", $text, -1);
			}
		}
		return $text;
	}
	
	
	function get_file($file, $method, $check_last_modified) { // method can be "http", "filesystem", or "auto" (auto looks at $pre_render_files list)
		GLOBAL $output, $file_missing, $pre_render_files, $script_location;
		GLOBAL $template, $read_page_definitions_from, $latest_mod_date, $latest_mod_date_formatted;
		GLOBAL $check_URL;
		$text = "";
		if (preg_match("/\/(\.\w.*)/", $file)) { // don't read files whose names start with "."
			$text = show_error("Not authorized to read the hidden file named \"$matches[1]\"");
		} else if (preg_match("/([^\w-\.\/\?\&\=\%])/", $file, $matches)) { // don't read files whose names contain characters that could be used to read remote files, or to hack into the open command (Perl only)
			$text = show_error("File names can't contain the character $matches[0]");
		} else {
			
			// first, figure out if we're going to read over HTTP or through the filesystem...
			if (($method == "")||($method == "auto")) {
				if ($pre_render_files) {
					for ($n=0; $n < count($pre_render_files); $n++) {
						if (strstr($file, $pre_render_files[$n])) { $method = "http"; }
					}
				}
			}
			
			// ...then get the file
			if ($method == "http") {

				// 1) prepare the URL for this type of connection
				$file_URL = absolute_path($file, "http");

				// in PHP, the next three steps are the same for http or local connections, but it's easier to maintain the
				// different versions of this script if we lay out the conditional nesting the same way in every language
				// hence the repeated code here
				
				// 2) open a connection to the file (in PHP, steps 2, 3 and 5 are shared by the http and filesystem options, so the nesting is different here)
				@$myfile = fopen($file_URL, "r");
				if (! $myfile) {
					$text = show_error("Can't open the file named \"$file\" $check_URL");
					$file_missing = 1;
				} else {
					// 3) read the file
					while (!feof($myfile)) { $text .= fread($myfile, 4096); }
					$file_missing = 0;
					
					// 4) close the file connection
					fclose($myfile);
					
				}
	
				//$output .= "<p>reading $file_URL over http<p>";  // debug

			} else {
			
				// 1) prepare the URL for this type of connection
				$file_URL = absolute_path($file, "filesystem");
				
				// 2) open a connection to the file (in PHP, steps 2, 3 and 5 are shared by the http and filesystem options, so the nesting is different here)
				@$myfile = fopen($file_URL, "r");
				if (! $myfile) {
					$text = show_error("Can't open the file named \"$file\" $check_URL");
					$file_missing = 1;
				} else {
				
					// 3) read the file
					while (!feof($myfile)) { $text .= fread($myfile, 4096); }
					$file_missing = 0;
					
					// 4) for local files, catch the file's modification date (for $read_page_definitions_from and $template, plus content files when the template contains a env_last_modified tag)
					if ($check_last_modified) {
						$mtime = filemtime($file_URL);
						if ($mtime > $latest_mod_date) {
							$latest_mod_date = $mtime;
							$latest_mod_date_formatted = date("Y-m-d - g:i a", $mtime); // h for two-digit hour
						}
					}
	
					// 5) close the file connection
					fclose($myfile);
					
				}
	
				//$output .= "<p>reading $script_location_copy + $file_copy through filesystem<p>"; // debug

			}
				
			// finally, conform all line breaks to avoid cross-platform issues
			// we'd rather conform to Unix line breaks (\n), but that doesn't work in ASP, which doesn't support the s flag in regular expressions
			$text = str_replace("\r\n", "\r", $text); // change DOS to Mac
			$text = str_replace("\n", "\r", $text); // change Unix to Mac

		}
		return $text;
	}
	
	function get_setting($source_text, $setting, $allow_override, $allow_partial) {
		// allow most preferences to be overridden through the argument string
		GLOBAL $output, $settings_used;
		$value = "";
		$setting_no_spaces = $setting;
		$setting_no_spaces = preg_replace("/[ -]/", "_", $setting_no_spaces, -1);
		$settings_used .= "$setting_no_spaces,";
		
		$override = (isset($_GET[$setting_no_spaces])) ? $_GET[$setting_no_spaces] : "" ;
		if (($allow_override)&&($override != "")) {
			$value = $override;
			// don't allow users to pass code through the URL arguments
			$value = str_replace("<", "&lt;", $value);
			$value = str_replace(">", "&gt;", $value);
		} else {
			$source_text = "\r".$source_text; // this makes it easy to make sure settings start on a new line
			if ($allow_partial) {  // this lets us use more verbose labels in the ini file
				if (preg_match("/[^\w]".$setting."[^:\r]*: ([^\r]*)\r*/i", $source_text, $matches)) {
					$value = $matches[1];
				}
			} else {  // this ensures an exact match in the pages file
				if (preg_match("/[^\w]".$setting.": ([^\r]*)\r*/i", $source_text, $matches)) {
					$value = $matches[1];
				}
			}
		}
		if ($value == "0") { $value = 0; } // this ensures that a 0 value will evaluate false in every language
		return $value;
	}
	
	function get_page_name($page_definition, $source) { // look for a page definition that exactly matches a given dynamic URL
		GLOBAL $output;
        $source = "\r".$source; // this makes it easier to match a page definition on the first line of the file
		$page_name = (preg_match("/[^\w]([-\w]+)\:\s+".preg_quote($page_definition, "/")."[\&\s\r]/", $source, $matches)) ? $matches[1] : "" ;
		return $page_name;
	}
	
	function get_page_names($source) { // return a list of all the page names in a page definitions file
		GLOBAL $output;
		$page_names = Array();
		while (preg_match("/([-\w]+)\:\s+[\w]+/", $source, $matches)) {
			$this_page_name = $matches[1];
			array_push($page_names, $this_page_name);
			$source = str_replace($this_page_name.":", "", $source);
		}
		return $page_names;
	}

	function get_directory_listing($directory) { // return a list of all the items in the specified directory
		GLOBAL $output;
		if ($dir = opendir(absolute_path($directory, "filesystem"))) {
			$files = Array();
			while ($file = readdir($dir)) {
				if (substr($file, 0, 1) != ".") { array_push($files, $file); } // that gets files that don't start with a dot
			}  
			closedir($dir);
		}
		return $files;
	}

	function absolute_path($path, $method) { // convert a reference to a file within our site to a complete filesystem path or absolute URL
		GLOBAL $output, $script_location, $request, $URL_rewriting_flag;

		if ($method == "filesystem") {
		
			if ($script_location == "NA") {
				$output .= show_error("Couldn't determine the filesystem path to this script. Please ensure that either the SCRIPT_FILENAME or PATH_TRANSLATED variable is available on your server.");
			} else {
				$script_location_copy = $script_location;
				$file_copy = $path;
				while (strstr($file_copy, "../")) {
					// this avoids results like contemplate/../templates/default.html
					$script_location_copy = preg_replace("/[^\/]+\/$/", "", $script_location_copy, 1); // remove the last directory name from the script location
					$file_copy = preg_replace("/\.\.\//", "", $file_copy, 1); // remove the first updelim from the filename (in PHP, we need to use preg_replace so we only do one replacement)
				}
				$path = $script_location_copy.$file_copy;
			}
			
		} else {
		
			if ($request == "NA") {
				$output .= show_error("Couldn't determine the web path to this script. Please ensure that either the SCRIPT_NAME or REQUEST_URI variable is available on your server.");
			} else {
				$thisserver = (isset($_SERVER["HTTP_HOST"])) ? $_SERVER["HTTP_HOST"] : "" ;
				$thispath = "";
				if ($URL_rewriting_flag) {
					if (preg_match("/\/([^=]*\/)$URL_rewriting_flag\//i", $request, $matches)) { $thispath = $matches[1]; }
				} else {
					if (preg_match("/\/([^=]*\/)contemplate\//i", $request, $matches)) { $thispath = $matches[1]; }
				}
				
				// pass along the authentication information if present
				$login = (isset($_SERVER["PHP_AUTH_USER"])) ? $_SERVER["PHP_AUTH_USER"].":".$_SERVER["PHP_AUTH_PW"]."@" : "" ; 
				
				$path = preg_replace("/\.\.\//", "http://$login$thisserver/$thispath", $path, 1); // replace the first updelim with the scheme and path (in PHP, we need to use preg_replace so we only do one replacement)
				$path = str_replace("../", "", $path); // remove additional updelims
			}
			
		}
		
		return $path;
	}

	function stri_replace($pattern, $replacement, $subject) { // do a case-insensitive replace; this is a substitute for a PHP function that doesn't exist!
		GLOBAL $output;
		return preg_replace("/".preg_quote($pattern, "/")."/is", $replacement, $subject, -1);
	}

	function show_error($message) { // error display isn't as well developed here as in assembler, but we'll put this here so we can share more code
		return "Error: ".$message;
	}
	
	
// set up some primary variables

	// try to figure out the path to this script as it appears in the browser, before any URL rewriting
	// in IIS, this probably won't work, because REQUEST_URI will either not exist, or will show the path after rewriting, since rewriting is handled externally from IIS
	$request = "NA";
	if (isset($_SERVER["REQUEST_URI"])) {
		$request = $_SERVER["REQUEST_URI"];
		if (strpos($request, "?") !== false) {
			$request = substr($request, 0, strpos($request, "?"));
		}
	} else {
		// IIS doesn't always provide the REQUEST_URI variable
		if (isset($_SERVER["SCRIPT_NAME"])) {
			$request = $_SERVER["SCRIPT_NAME"];
		}
	}
	
	// we don't do URL rewriting except in the assembler, so this value should be empty
	// also, we can't tell what flag(s) the assembler is using, but we can set a default here
	$URL_rewriting_flag = "";
	$URL_rewriting_flag_default = "assembled";

	// figure out the full local path of the contemplate directory so we can build an absolute path that works on Windows computers
	// on Unix, this will include any symlinks used to construct the URL, so we can run the script from a symlink to the real contemplate folder, 
	// and paths will be relative to the apparent, rather than the actual, location of assembler.asp
	// on Windows, this will not include virtual directories, but constructing a path from a virtual directory wouldn't work anyway
	// because those aren't available at the file system level
	$script_location = "NA";
	if (isset($_SERVER["SCRIPT_FILENAME"])) {
		$script_location = $_SERVER["SCRIPT_FILENAME"];
	} else {
		// Windows doesn't provide the SCRIPT_FILENAME variable
		if (isset($_SERVER["PATH_TRANSLATED"])) {
			$script_location = $_SERVER["PATH_TRANSLATED"];
		}
	}
	$script_location = str_replace("\\", "/", $script_location);
	$script_location = str_replace("//", "/", $script_location); // not sure why PHP returns double slashes on Windows
	if (($URL_rewriting_flag)&&(preg_match("/(.*\/)$URL_rewriting_flag\//i", $script_location, $matches))) {
		$script_location = $matches[1]."contemplate/";
	} else {
		$script_location = preg_replace("/(.*\/)[^\/]+/", "$1", $script_location, 1);
	}
	
	$only = (isset($_POST["only"])) ? $_POST["only"] : "" ;
	$skip = (isset($_POST["skip"])) ? $_POST["skip"] : "" ;
	$mode = (isset($_POST["mode"])) ? $_POST["mode"] : "" ;
	$flatten = (isset($_POST["flatten"])) ? $_POST["flatten"] : "" ;
	$input_file = (isset($_POST["input_file"])) ? $_POST["input_file"] : "" ;
	$output_file = (isset($_POST["output_file"])) ? $_POST["output_file"] : "" ;
		
	$processed_count = 0;
	$skipped_count = 0;
	$error_count = 0;
	$changes = 0;


// read in a preferences (.ini) file if one exists

	$preferences = get_file("../data/preferences.ini", "filesystem", 0);
	if ($preferences) {
		$field_marker = get_setting($preferences, "field marker", 0, 1);
		$read_page_definitions_from = "../".get_setting($preferences, "read page definitions from", 0, 1);
	} else {
		$field_marker = "";
		$read_page_definitions_from = "";
	}
	
	
// do a bit of processing on the settings, and set some defaults in case preferences.ini or individual settings are absent

	if ($field_marker == "") { $field_marker = "##"; }

	if ($read_page_definitions_from == "") { $read_page_definitions_from = "../data/pages.txt"; }


// set up the HTML page
	
	print "
	
		<html>
	
		<head>
		
			<title>Contemplate(tm) Web Templating System</title>
			
			<link rel=stylesheet href=../stylesheet.css type=text/css>
		
		</head>
		
		<body bgcolor=#FFFFFF>
				
			<p class=title>Contemplate</p>
			<p class=body><a href=../index.html>Utilities menu</a> &gt; Flattener</p><br>
	
			<div class=body>
			
	";
	

// require a current version of the assembler before proceeding

	$required_version = 1.0; // we assume that any sub-sub-version will be compatible
	$extractor_text = get_file("../assembler.php", "filesystem", 0);
	
	if (preg_match("/version (\d+\.\d+)/", $extractor_text, $matches)) {
		$installed_version = $matches[1];
	} else {
		$installed_version = 0;
	}
	if ($installed_version < $required_version) {
		print "Error: This version of the Flattener requires version $required_version or newer of Contemplate (assembler.php).<p>";
		$stop = 1;
	}
	

// present some options on the first run
	
	if ((!$flatten)&&(!$input_file)&&(!$stop)) {

	/////
	// show the intro page
	/////
	
		print "
			
			<p class=subheading>Flatten assembled pages</p>
			<p class=body>Before you do this, create a folder called \"flattened\" at the root level of your website, and give it world-write access (chmod a+w flattened)</p>
			
			<form name=flatten method=post>
			
				<table><tr valign=bottom><td class=body>
				
					Only names containing...<br>
					<input type=text name=only size=32>
				
				</td><td class=body>
				
					Skip names containing...<br>
					<input type=text name=skip size=32>
				
				</td></tr></table><br>
				
				<p class=body>
					<input type=radio name=mode value=flatten checked> Flatten only &nbsp;
					<input type=radio name=mode value=preview> Preview only &nbsp;
					<input type=radio name=mode value=both> Flatten and preview
				</p>
				
				<p class=body><input type=submit value=Flatten></p><br>
				
				<input type=hidden name=flatten value=1>
				
			</form>
			
			<p class=subheading>Convert links in a static page</p>
			<p class=body>If you have a static HTML page that contains links to assembled pages, this option will convert the links to use the flattened versions of those pages</p>
			
			<form name=parse method=post>
			
				<table><tr valign=bottom><td class=body>
				
					Input file (e.g. index.html)<br>
					<input type=text name=input_file size=32>
				
				</td><td class=body>
				
					Output file (e.g. flattened/index.html)<br>
					<input type=text name=output_file size=32>
				
				</td></tr></table><br>
				
				<p class=body><input type=submit value=Convert></p><br>
			
			</form>
		
		";
		

	} else {
	
	
	// grab the text of the page definitions file; both options will use it
	
		$pages_file = get_file($read_page_definitions_from, "filesystem", 0);
		if ($file_missing) { print $pages_file."<p>"; } // the result of get_file will contain the error message
		
	
		if (($flatten)&&(!$file_missing)) {
		
		////
		// flatten assembled pages
		////
		
			// parse the pages file to get an array of page names and an array of URLs
			
				$pages = Array(); $URLs = Array();
				$page_names = get_page_names($pages_file);
				for ($i=0; $i<count($page_names); $i++) {
					$this_page_name = $page_names[$i];
					$this_page = "../flattened/$this_page_name.html";
					$this_URL = "../contemplate/assembler.php?page=$this_page_name&skip_rendering=1";
					$pages[$i] = $this_page;
					$URLs[$i] = $this_URL;
				}
			
			// open each URL, change its dynamic links to static ones, and save it to its page name
			
				$verb_present = ($mode == "preview") ? "Previewing" : "Flattening" ;
				$verb_past = ($mode == "preview") ? "Previewed" : "Flattened" ;

				for ($i=0; $i<count($pages); $i++) {
				
					// skip some pages if the user put something into the filter fields
					if (($only)&&(!stristr($pages[$i], $only))) {
						$result .= "Skipping $pages[$i]<br>";
						$skipped_count++;
					} else if (($skip)&&(stristr($pages[$i], $skip))) {
						$result .= "Skipping $pages[$i]<br>";
						$skipped_count++;
					} else {
						$processed_count++;
				
						// get the dynamic page
						$file_missing = 0;
						$page_rendered = get_file("../".$URLs[$i], "http", 0);
						
						if ($file_missing) {
							$result .= "<font color=red>$page_rendered</font><br>"; // that will contain the error if $file_missing == 1
							$error_count++;
						} else {
						
							// convert all the dynamic URLs to flattened ones
							$page_rendered = convert_URLs($page_rendered, $pages[$i]);
							
							// if a flatten is requested, save the flattened HTML
							if (($mode == "flatten")||($mode == "both")) {
								// leave the next line commented out to use relative paths and avoid a bug with PHP and open_basedir before version 4.3 (this appears in two other places in this file)
								//$pages[$i] = absolute_path($pages[$i], "filesystem");
								@$my_file = fopen("../".$pages[$i], "w");
								if ($my_file) {
									fwrite ($my_file, $page_rendered);
									$result .= "$verb_present $pages[$i]<br>";
									fclose ($my_file);
								} else {
									$result .= "<font color=red>Error: Can't open the file $pages[$i] for flattening</font><br>";
									$error_count++;
								}
							}
								
							// if a preview is requested, show the flattened HTML
							if (($mode == "preview")||($mode == "both")) {
								if (preg_match("/<body[^>]*>(.*)<\/body>/i", $page_rendered, $matches)) {
									$page_preview = $matches[1];
									$page_preview = preg_replace("/([^\/])\.\.\//", "$1../../", $page_preview); // add another level to relative paths so that graphics and links still work
									if (($mode == "preview")||(!$my_file)) {
										// if we didn't show a flatten label, or flattening failed, show a preview label
										$result .= "Previewing $pages[$i]<br>";
									}
									$result .= "<table border=1><tr><td>$page_preview</td></table><br>";
								} else {
									$result .= "Error: Can't display a preview for the file $pages[$i]<br>";
									$error_count++;
								}
							}

						}
					
					}
					
					// provide feedback as the script runs
					print $result;
					$result = "";
				
				}
		
			// show success rate and, if errors, offer tips
		
				print "<br><b>Processed $processed_count files, skipped $skipped_count files, and encountered $error_count errors</b>";
		
	
		} elseif ($input_file) {
	
		/////
		// convert links in a static page
		////
		
			// open the file, change its dynamic links to static ones, and save it to its output location
			
				// in case the paths aren't relative to the contemplate folder
				$input_file = stristr($input_file, "../") ? $input_file : "../".$input_file ;
				$output_file = stristr($output_file, "../") ? $output_file : "../".$output_file ;
				
				// get the file
				$delimiter = (!strstr($input_file, "?")) ? "?" : "&" ;
				$file = get_file("../".$input_file.$delimiter."skip_rendering=1", "http", 0);

				// convert all the dynamic URLs to flattened ones
				$file = convert_URLs($file, $input_file);
			
				// save the parsed file
				// leave the next line commented out to use relative paths and avoid a bug with PHP and open_basedir before version 4.3 (this appears in two other places in this file)
				//$output_file = absolute_path($output_file, "filesystem");
				$my_file = fopen("../".$output_file, "w");
				if (!$my_file) {
					$error_count++;
				} else {
					fwrite($my_file, $file);
					fclose ($my_file);
				}
				
			// show result
		
				if (!$error_count) {
					print "<b>Converted $input_file successfully</b><p>";
				} else {
					print "Error: Couldn't open the file $output_file for saving.<br>";
					print "<b>Failed to convert $input_file</b><p>";
				}
				
		
		}
	
		if (!$stop) {
			print "<p class=body><a href=JavaScript:history.back()>Back</a></p>";
		}

		print $output; // we don't use the $output variable in this script, but this helps us share code with assembler.php
				

// finish off a big conditional up there
	
	}	
	
	
// we're done!

	print "</div>";

	print "</body></html>";

?>
