<?php


// get variables passed in from do_save()

	$pages_file_new = (isset($_POST["pages_file"])) ? $_POST["pages_file"] : "" ;
	$action = (isset($_POST["action"])) ? $_POST["action"] : "" ;


// define some useful functions

	function get_file($file, $method, $check_last_modified) { // method can be "http", "filesystem", or "auto" (auto looks at $pre_render_files list)
		GLOBAL $output, $file_missing, $pre_render_files, $script_location;
		GLOBAL $template, $read_page_definitions_from, $latest_mod_date, $latest_mod_date_formatted;
		GLOBAL $check_URL;
		$text = "";
		if (preg_match("/\/(\.\w.*)/", $file)) { // don't read files whose names start with "."
			$text = show_error("Not authorized to read the hidden file named \"$matches[1]\"");
		} else if (preg_match("/([^\w-\.\/\?\&\=\%])/", $file, $matches)) { // don't read files whose names contain characters that could be used to read remote files, or to hack into the open command (Perl only)
			$text = show_error("File names can't contain the character $matches[0]");
		} else {
			
			// first, figure out if we're going to read over HTTP or through the filesystem...
			if (($method == "")||($method == "auto")) {
				if ($pre_render_files) {
					for ($n=0; $n < count($pre_render_files); $n++) {
						if (strstr($file, $pre_render_files[$n])) { $method = "http"; }
					}
				}
			}
			
			// ...then get the file
			if ($method == "http") {

				// 1) prepare the URL for this type of connection
				$file_URL = absolute_path($file, "http");

				// in PHP, the next three steps are the same for http or local connections, but it's easier to maintain the
				// different versions of this script if we lay out the conditional nesting the same way in every language
				// hence the repeated code here
				
				// 2) open a connection to the file (in PHP, steps 2, 3 and 5 are shared by the http and filesystem options, so the nesting is different here)
				@$myfile = fopen($file_URL, "r");
				if (! $myfile) {
					$text = show_error("Can't open the file named \"$file\" $check_URL");
					$file_missing = 1;
				} else {
				
					// 3) read the file
					while (!feof($myfile)) { $text .= fread($myfile, 4096); }
					$file_missing = 0;
					
					// 4) close the file connection
					fclose($myfile);
					
				}
	
				//$output .= "<p>reading $file_URL over http<p>";  // debug

			} else {

				// 1) prepare the URL for this type of connection
				$file_URL = absolute_path($file, "filesystem");
				
				// 2) open a connection to the file (in PHP, steps 2, 3 and 5 are shared by the http and filesystem options, so the nesting is different here)
				@$myfile = fopen($file_URL, "r");
				if (! $myfile) {
					$text = show_error("Can't open the file named \"$file\" $check_URL");
					$file_missing = 1;
				} else {
				
					// 3) read the file
					while (!feof($myfile)) { $text .= fread($myfile, 4096); }
					$file_missing = 0;
					
					// 4) for local files, catch the file's modification date (for $read_page_definitions_from and $template, plus content files when the template contains a env_last_modified tag)
					if ($check_last_modified) {
						$mtime = filemtime($file_URL);
						if ($mtime > $latest_mod_date) {
							$latest_mod_date = $mtime;
							$latest_mod_date_formatted = date("Y-m-d - g:i a", $mtime); // h for two-digit hour
						}
					}
	
					// 5) close the file connection
					fclose($myfile);
					
				}
	
				//$output .= "<p>reading $script_location_copy + $file_copy through filesystem<p>"; // debug

			}
				
			// finally, conform all line breaks to avoid cross-platform issues
			// we'd rather conform to Unix line breaks (\n), but that doesn't work in ASP, which doesn't support the s flag in regular expressions
			$text = str_replace("\r\n", "\r", $text); // change DOS to Mac
			$text = str_replace("\n", "\r", $text); // change Unix to Mac

		}
		return $text;
	}
	
	function get_setting($source_text, $setting, $allow_override, $allow_partial) {
		// allow most preferences to be overridden through the argument string
		GLOBAL $output, $settings_used;
		$value = "";
		$setting_no_spaces = $setting;
		$setting_no_spaces = preg_replace("/[ -]/", "_", $setting_no_spaces, -1);
		$settings_used .= "$setting_no_spaces,";
		
		$override = (isset($_GET[$setting_no_spaces])) ? $_GET[$setting_no_spaces] : "" ;
		if (($allow_override)&&($override != "")) {
			$value = $override;
			// don't allow users to pass code through the URL arguments
			$value = str_replace("<", "&lt;", $value);
			$value = str_replace(">", "&gt;", $value);
		} else {
			$source_text = "\r".$source_text; // this makes it easy to make sure settings start on a new line
			if ($allow_partial) {  // this lets us use more verbose labels in the ini file
				if (preg_match("/[^\w]".$setting."[^:\r]*: ([^\r]*)\r*/i", $source_text, $matches)) {
					$value = $matches[1];
				}
			} else {  // this ensures an exact match in the pages file
				if (preg_match("/[^\w]".$setting.": ([^\r]*)\r*/i", $source_text, $matches)) {
					$value = $matches[1];
				}
			}
		}
		if ($value == "0") { $value = 0; } // this ensures that a 0 value will evaluate false in every language
		return $value;
	}
	
	function get_page_names($source) { // return a list of all the page names in a page definitions file
		GLOBAL $output;
		$page_names = Array();
		while (preg_match("/([-\w]+)\:\s+[\w]+/", $source, $matches)) {
			$this_page_name = $matches[1];
			array_push($page_names, $this_page_name);
			$source = str_replace($this_page_name.":", "", $source);
		}
		return $page_names;
	}

	function get_directory_listing($directory) { // return a list of all the items in the specified directory
		GLOBAL $output;
		if ($dir = opendir(absolute_path($directory, "filesystem"))) {
			$files = Array();
			while ($file = readdir($dir)) {
				if (substr($file, 0, 1) != ".") { array_push($files, $file); } // that gets files that don't start with a dot
			}  
			closedir($dir);
		}
		return $files;
	}
	
	function get_recursive_listing($start_folder, $recursive=0, $filter="") { // return a list of all the files in a given folder, including subfolders
		GLOBAL $output;
		$start_folder = preg_replace("/[\\/\\\]$/", "", $start_folder, 1); // remove a trailing slash
		$file_list = Array();
		$folder_list = Array($start_folder);
		while ($folder_list) {
			$this_folder_path = $folder_list[0];
			$this_folder_items = get_directory_listing($this_folder_path);
			for ($n=0; $n<count($this_folder_items); $n++) {
				$this_item_name = $this_folder_items[$n];
				$this_item_path = $this_folder_path."/".$this_item_name;
 				// if the item name contains a dot, we assume it's a file, if not, we assume it's a directory
 				if (! strstr($this_item_name, ".")) {
 					// we've found a folder; add it to the list of folders to scan
 					if ($recursive) {
						array_push($folder_list, $this_item_path);
					}
				} else if ((!$filter)||(strstr($this_item_name, $filter))) {
					// we've found a file; add it to the list of files to return
					$this_item_path = str_replace($start_folder."/", "", $this_item_path); // make results relative to the start folder, not to the site
					array_push($file_list, $this_item_path);
				}
			}
			array_shift($folder_list); // remove the folder we just scanned from the list
		}
		return $file_list;
	}
	
	function absolute_path($path, $method) { // convert a reference to a file within our site to a complete filesystem path or absolute URL
		GLOBAL $output, $script_location, $request, $URL_rewriting_flag;

		if ($method == "filesystem") {
		
			if ($script_location == "NA") {
				$output .= show_error("Couldn't determine the filesystem path to this script. Please ensure that either the SCRIPT_FILENAME or PATH_TRANSLATED variable is available on your server.");
			} else {
				$script_location_copy = $script_location;
				$file_copy = $path;
				while (strstr($file_copy, "../")) {
					// this avoids results like contemplate/../templates/default.html
					$script_location_copy = preg_replace("/[^\/]+\/$/", "", $script_location_copy, 1); // remove the last directory name from the script location
					$file_copy = preg_replace("/\.\.\//", "", $file_copy, 1); // remove the first updelim from the filename (in PHP, we need to use preg_replace so we only do one replacement)
				}
				$path = $script_location_copy.$file_copy;
			}
			
		} else {
		
			if ($request == "NA") {
				$output .= show_error("Couldn't determine the web path to this script. Please ensure that either the SCRIPT_NAME or REQUEST_URI variable is available on your server.");
			} else {
				$thisserver = (isset($_SERVER["HTTP_HOST"])) ? $_SERVER["HTTP_HOST"] : "" ;
				$thispath = "";
				if ($URL_rewriting_flag) {
					if (preg_match("/\/([^=]*\/)$URL_rewriting_flag\//i", $request, $matches)) { $thispath = $matches[1]; }
				} else {
					if (preg_match("/\/([^=]*\/)contemplate\//i", $request, $matches)) { $thispath = $matches[1]; }
				}
				
				// pass along the authentication information if present
				$login = (isset($_SERVER["PHP_AUTH_USER"])) ? $_SERVER["PHP_AUTH_USER"].":".$_SERVER["PHP_AUTH_PW"]."@" : "" ; 
				
				$path = preg_replace("/\.\.\//", "http://$login$thisserver/$thispath", $path, 1); // replace the first updelim with the scheme and path (in PHP, we need to use preg_replace so we only do one replacement)
				$path = str_replace("../", "", $path); // remove additional updelims
			}
			
		}
		
		return $path;
	}

	function show_error($message) { // the only purpose for this here is to prevent the script from halting if one of the shared functions throws an error
		return "Error: ".$message;
	}
	

// set up some error messages that we can use later

	$check_URL = "<br>Hint: Check your page definition";
	$check_embed = "<br>Hint: Check your embed tag";
	$check_field_label = "<br>Hint: Check your content file and remove any formatting tags between the field marker and the field label";
	

// set up some primary variables

	// try to figure out the path to this script as it appears in the browser, before any URL rewriting
	// in IIS, this probably won't work, because REQUEST_URI will either not exist, or will show the path after rewriting, since rewriting is handled externally from IIS
	$request = "NA";
	if (isset($_SERVER["REQUEST_URI"])) {
		$request = $_SERVER["REQUEST_URI"];
		if (strpos($request, "?") !== false) {
			$request = substr($request, 0, strpos($request, "?"));
		}
	} else {
		// IIS doesn't always provide the REQUEST_URI variable
		if (isset($_SERVER["SCRIPT_NAME"])) {
			$request = $_SERVER["SCRIPT_NAME"];
		}
	}
	
	// we don't do URL rewriting except in the assembler, so this value should be empty
	// also, we can't tell what flag(s) the assembler is using, but we can set a default here
	$URL_rewriting_flag = "";
	$URL_rewriting_flag_default = "assembled";

	// figure out the full local path of the contemplate directory so we can build an absolute path that works on Windows computers
	// on Unix, this will include any symlinks used to construct the URL, so we can run the script from a symlink to the real contemplate folder, 
	// and paths will be relative to the apparent, rather than the actual, location of assembler.asp
	// on Windows, this will not include virtual directories, but constructing a path from a virtual directory wouldn't work anyway
	// because those aren't available at the file system level
	$script_location = "NA";
	if (isset($_SERVER["SCRIPT_FILENAME"])) {
		$script_location = $_SERVER["SCRIPT_FILENAME"];
	} else {
		// Windows doesn't provide the SCRIPT_FILENAME variable
		if (isset($_SERVER["PATH_TRANSLATED"])) {
			$script_location = $_SERVER["PATH_TRANSLATED"];
		}
	}
	$script_location = str_replace("\\", "/", $script_location);
	$script_location = str_replace("//", "/", $script_location); // not sure why PHP returns double slashes on Windows
	if (($URL_rewriting_flag)&&(preg_match("/(.*\/)$URL_rewriting_flag\//i", $script_location, $matches))) {
		$script_location = $matches[1]."contemplate/";
	} else {
		$script_location = preg_replace("/(.*\/)[^\/]+/", "$1", $script_location, 1);
	}


// get some preference settings from the preferences file

	$preferences = get_file("../data/preferences.ini", "filesystem", 0);
	if ($preferences) { $read_page_definitions_from = "../".get_setting($preferences, "read page definitions from", 0, 1); }
	if (($read_page_definitions_from == "../")||($read_page_definitions_from == "")) { $read_page_definitions_from = "../data/pages.txt"; }


// set up the HTML page
	
	print "
	
		<html>
	
		<head>
		
			<title>Contemplate</title>
			
			<link rel=stylesheet href=../stylesheet.css type=text/css>
		
		</head>
		
		<body bgcolor=#FFFFFF>
				
			<div class=body>
	
	";
	

// try to update the page definitions file

	$error = 0;

	// first find out if the page definitions file has been locked by Dreamweaver
	
	$lock_flag = get_file("$read_page_definitions_from.lck", "filesystem", 0);
	if (! $file_missing) {
		$error = 1;
		print "Error: The file $pages_file is currently checked out.<br>";
	}

	if (! $error) {

		// then read the existing file and back it up
		
		$pages_file = get_file($read_page_definitions_from, "filesystem", 0);
		$save_definitions_backup_to = $read_page_definitions_from.".backup";
		$myfile = @fopen($save_definitions_backup_to, "w");
		if (! $myfile) {
			$error = 1;
			print "Error: Can't open the file $save_definitions_backup_to for backup.<br>";
		} else {
			fputs($myfile, $pages_file);
			fclose($myfile);
		}
	
	}
	
	if (! $error) {
	
		// now save the new definitions to the original file
	
		$myfile = @fopen($read_page_definitions_from, "w");
		if (! $myfile) {
			$error = 1;
			print "Error: Can't open the file $read_page_definitions_from for updating.<br>";
		} else {
			fputs($myfile, $pages_file_new);
			fclose($myfile);
		}
		
	}
	
	
// if all went well, print a message for the user

	if (! $error) {
	
		if ($action == "delete") {
		
			print "Your old page definition has been deleted!";

		} else {

			print "
			
				<script language=JavaScript>
					relative_URL = 'contemplate/assembler.php?page=' + parent.page_name;
					site_root = document.URL.substr(0, document.URL.indexOf('/contemplate/formulator/'));
					absolute_URL = site_root + '/' + relative_URL;
					relative_URL = '../' + relative_URL;
				</script>
	
				Your new page definition has been saved!
				
				<script language=JavaScript>
					document.write('<a href=' + absolute_URL + ' target=_new>Click here</a> to test it.<p>');
				</script>
	
				You can link to your new page from within your site at...
				
				<blockquote>
				
					<script language=JavaScript>
						document.write(relative_URL);
					</script>
				
				</blockquote>
			
				...or from other documents at...
				
				<blockquote>
				
					<script language=JavaScript>
						document.write(absolute_URL);
					</script>
				
				</blockquote>
			
			";
			
		}
	
	}


// and write the new page definitions into a hidden field, in case we want to save again from this page

	$pages_file_new = str_replace("\"", "'", $pages_file_new); // blech ... change double to single quotes so they don't interfere with the input tag quotes
	
	print "
	
		<form name=contemplate method=post action=save.php>
	
			<input type=hidden name=pages_file value=\"$pages_file_new\">
			
		</form>

	";
	
	
// we're done!

	print "</div>";

	print "</body></html>";

	exit;
